import pandas as pd
import numpy as np
import plotly.express as px
import re
from multiprocessing import Pool
import unidecode

import warnings

from pandas.api.types import is_float_dtype, is_numeric_dtype, is_string_dtype
from humailib.cloud_tools import GoogleBigQuery, GoogleCloudStorage, AzureBlobStorage, AzureDataLake

def instantiate_class(module_name, class_name):
    module = __import__(module_name, globals(), locals(), [class_name])
    class_ = getattr(module, class_name)
    
    return class_


keys = {
    'azure' : {
        'gcs' : 'DefaultEndpointsProtocol=https;AccountName=humaimain;AccountKey=VSWuKEKDm3FC9dV43g6giU2RD93XkMyZ4OrXiBqOdu/nHozxlZjTmZcOyg7D2LzI/qjGSvanMrouG9fzhDOPVw==;EndpointSuffix=core.windows.net',
        'gbq_humai_sb' : 'DefaultEndpointsProtocol=https;AccountName=humaiadlsgen2;AccountKey=1awKbHfeHcqWYtGut/t2t8dBM0ZLXCWlyVUePE+EuqXOzAXeDwF8AOb9MvHvkexuPNDmfAJt8iab80Xp97lM9Q==;EndpointSuffix=core.windows.net',
        'gbq_humai_insights' : 'DefaultEndpointsProtocol=https;AccountName=humaiadlsgen2;AccountKey=1awKbHfeHcqWYtGut/t2t8dBM0ZLXCWlyVUePE+EuqXOzAXeDwF8AOb9MvHvkexuPNDmfAJt8iab80Xp97lM9Q==;EndpointSuffix=core.windows.net',
    },
    'gcp' : {
        'gcs' : 'service_account_keys/humai-sb-7b9db70de787.json',
        'gbq_humai_sb' : 'service_account_keys/humai-sb-7b9db70de787.json',
        'gbq_humai_insights' : 'service_account_keys/humai-insights-78e122a483b7.json',
    }
}

def get_io(platform, path='../../'):

    if platform.upper() == 'AZURE':
        print("[get_io] Microsoft Azure...")
        key = keys['azure']
        gcs = AzureBlobStorage(key=key['gcs'])
        gbq_humai_sb = AzureDataLake(project='humai-sb', key=key['gbq_humai_sb'])
        gbq_humai_insights = AzureDataLake(project='humai-insights', key=key['gbq_humai_insights'])
    else:
        print("[get_io] Google Cloud Platform...")
        key = keys['gcp']
        gcs = GoogleCloudStorage(key=path+key['gcs'])
        gbq_humai_sb = GoogleBigQuery(project='humai-sb', key=path+key['gbq_humai_sb'])
        gbq_humai_insights = GoogleBigQuery(project='humai-insights', key=path+key['gbq_humai_insights'])

    return gcs, gbq_humai_sb, gbq_humai_insights


"""
Column type definitions from the data dictionary.
    
https://humai.sharepoint.com/:w:/s/Platform/ET6ZzN-QgiJIpLdbVMkFz0oBe9_UOhSoau0Zvbo7HJ-5lQ?e=OxMlYv
"""
humai_table_field_types = {    
    'transactions' : {
        'datetime' : ['Datetime'],
        'string' : ['Transaction_ID','Branch_ID','Customer_ID','Status','Single_Item_Product_ID','Zip','Item_Types'],
        'float' : ['Total_Gross','Discount','Amount_Paid','Voucher_Amount'],
        'int' : ['Total_Num_Items_Purchased'],
        'bool' : ['Voucher_Used'],
    },
    'items' : {
        'datetime' : ['Datetime'],
        'string' : ['Transaction_ID','Product_ID','Name','Item_Type'],
        'float' : ['Price', 'Discount'],
        'int' : ['Quantity'],
    },
    'mailings' : {
        'datetime' : ['Send_Datetime'],
        'string' : ['Broadcast_ID','Campaign_ID'],
        'int' : ['Num_Customers']
    },
    'mailings_events' : {
        'datetime' : ['Activity_Datetime', 'Send_Datetime'],
        'string' : ['Broadcast_ID','Customer_ID','Activity','Description'],
    },
    'customers' : {
        'string' : ['Customer_ID','Email'],
    },
    'customer_ids' : {
        'string' : ['Customer_ID','Personifier','Person']
    },
    'products' : {
        'string' : ['Name','Product_ID','SKU','Item_Type'],
        'float' : ['Price'],
    },
    'customer_characteristics' : {
        'string' : ['Customer_ID']
    },
    'product_characteristics' : {
        'string' : ['Product_ID']
    },
    'predictions' : {
        'string' : ['Customer_ID'],
        'datetime' : ['Date'],
        #'float' : ['CLV'],
        #'int' : ['Time_Of_Day']
    },
    'customer_lifetime_value' : {
        'float' : ['CLV'],
        'string' : ['Customer_ID'],
    },
}

def transactions_check_basic_assumptions(df):
    
    assert len(df[df.Amount_Paid < 0]) == 0
    assert len(df[df.Total_Gross < 0]) == 0
    assert len(df[df.Discount > 0]) == 0
    assert len(df) == df.Transaction_ID.nunique()
    assert len(df[df.Customer_ID == '']) == 0
    assert len(df[df.Transaction_ID == '']) == 0
    assert len(df[df.Total_Gross < df.Amount_Paid]) == 0
    
    assert len(df[df.Total_Num_Items_Purchased <= 0]) == 0
    
    
def items_check_basic_assumptions(df):
    
    assert len(df[df.Product_ID == '']) == 0
    assert len(df[df.Price < 0]) == 0
    assert len(df[df.Discount > 0]) == 0
    assert len(df[df.Quantity <= 0]) == 0
    

def customers_check_basic_assumptions(df):

    assert len(df) == df.Customer_ID.nunique()
    assert len(df[df.Customer_ID == '']) == 0
    
    
def products_check_basic_assumptions(df):
    
    assert len(df[df.Product_ID == '']) == 0
    assert len(df) == df.Product_ID.nunique()
    assert len(df[df.Price < 0]) == 0
    
    
humai_table_checks = {
    'transactions' : transactions_check_basic_assumptions,
    'items' : items_check_basic_assumptions,
    'customers' : customers_check_basic_assumptions,
    'products' : products_check_basic_assumptions,
}


def check_table(df, table_type, hard_assert=True):
    
    print("Checking columns...")
    for _,cols in humai_table_field_types.get(table_type,[]).items():
        for c in cols:
            if c in df:
                print("  '{}' -> Exists".format(c))
            else:
                if hard_assert:
                    assert c in df, "  '{}' -> Missing!".format(c)
                else:
                    warnings.warn(
                        "  '{}' -> Missing!".format(c),
                        category=UserWarning, 
                        stacklevel=1
                    )
         
    if table_type in humai_table_checks:
        print("Checking basic assumptions...")
        humai_table_checks[table_type](df)
                
                
def percentage_match_stats(df1, col1, df2, col2):
    vals1 = set(df1[col1].to_numpy())
    vals2 = set(df2[col2].to_numpy())
    
    print(f"Number of unique '{col1}': {len(vals1)}, unique '{col2}': {len(vals2)}")
    prop = 100.0 * len(vals1.intersection(vals2)) / len(vals1)
    print("Proportion of '{}' found in '{}': {:.2f} %".format(col1, col2, prop))
    
    diff = list(vals1.difference(vals2))[:5]
    print("First 5 of the difference set: {}".format(diff))
    

def convert_table_columns(
    df, 
    table_type = 'transactions', 
    datetime_format='%Y-%m-%d %H:%M:%S'
):
    
    """
    Convert table column types to table column type definitions as specified in the data dictionary:
    
    https://humai.sharepoint.com/:w:/s/Platform/ET6ZzN-QgiJIpLdbVMkFz0oBe9_UOhSoau0Zvbo7HJ-5lQ?e=OxMlYv
    """
    
    types = humai_table_field_types[table_type]
    if 'datetime' in types:
        columns_as_datetime(df, columns=types['datetime'], datetime_format=datetime_format)
    if 'string' in types:
        columns_as_str(df, columns=types['string'], uppercase=False, drop_empty=False)
    if 'float' in types:
        columns_as_float(df, columns=types['float'], na_replace_value=None, drop_na=False)
    if 'int' in types:
        columns_as_int(df, columns=types['int'], na_replace_value=None, drop_na=False)

    for c in df.columns: print("  {} -> {}".format(c, df[c].dtype))
        
        
def load_table(
    gbq, 
    dataset_table,
    table_type,
    datetime_columns = None,
    datetime_format = '%Y-%m-%d %H:%M:%S', 
    check_columns = True,
    flush_cache = True,
    hard_assert = False,
    use_pandas_gbq = True,
):
    """
    Load table and convert column types
    """    
         
    print("[load_table] Loading table...")
    df = gbq.download_table_to_pandas(dataset_table, flush_cache=flush_cache, use_pandas_gbq=use_pandas_gbq)

    if df is None:
        if hard_assert:
            assert df is not None, "  Could not load '{}".format(dataset_table)
        else:
            warnings.warn(
                "  Could not load '{}".format(dataset_table),
                category=UserWarning, 
                stacklevel=1
            )
        return None
    
    print("[load_table] Converting table columns...")
    if datetime_columns is not None:
        columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
      
    
    if table_type is not None:
        
        do_continue = True
        if check_columns:
            for _,columns in humai_table_field_types[table_type].items():
                for c in columns:
                    if c not in df:
                        warnings.warn(
                            f"[load_table] Column {c} is missing...",
                            category=UserWarning, 
                            stacklevel=1
                        )
                        do_continue = False

        if not do_continue:
            return None

        convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)    

        if table_type == 'transactions' and keep_most_recent_trans_id:
            print("[load_table] Keeping only the most recent Transaction ID, if there's multiples...")
            print("[load_table]   {:,} rows total.".format(len(df)))
            keep_most_recent_transaction_id(df)
            print("[load_table]   {:,} rows left.".format(len(df)))

    return df


def load_csv(
    gcs,
    filename,
    table_type,
    datetime_columns = None,
    datetime_format='%Y-%m-%d %H:%M:%S',
    check_columns=True,
    flush_cache = True,
    hard_assert = False,
):
    """
    Load CSV from GCS and convert column types.
    """
    
    print("[load_table] Loading table...")
    df = gcs.download_csv(filename, flush_cache=flush_cache)
    
    if df is None:
        if hard_assert:
            assert df is not None, "  Could not load '{}".format(filename)
        else:
            warnings.warn(
                "  Could not load '{}".format(filename),
                category=UserWarning, 
                stacklevel=1
            )
        return None
    
    print("[load_table] Converting table columns...")
    if datetime_columns is not None:
        columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
        
    if table_type is not None:
        
        do_continue = True
        if check_columns:
            for _,columns in humai_table_field_types[table_type].items():
                for c in columns:
                    if c not in df:
                        warnings.warn(
                            f"[load_table] Column {c} is missing...",
                            category=UserWarning, 
                            stacklevel=1
                        )
                        do_continue = False

        if not do_continue:
            return None
        
        convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)
    
        if table_type == 'transactions' and keep_most_recent_trans_id:
            print("[load_csv] Keeping only the most recent Transaction ID, if there's multiples...")
            print("[load_csv]   {:,} rows total.".format(len(df)))
            keep_most_recent_transaction_id(df)
            print("[load_csv]   {:,} rows left.".format(len(df)))

    return df


def load_and_merge_tables(
    gbq, 
    dataset,
    table_type,
    date_ranges,
    table_name = None,
    datetime_columns = None,
    datetime_format = '%Y-%m-%d %H:%M:%S', 
    check_columns = True,
    keep_most_recent_trans_id=True,
    flush_cache = True,
    hard_assert = False,
):
    """
    1. Load tables by date range. 
    2. Remove duplicate rows.
    3. Convert column types
    """    
    
    print("[load_and_merge_tables] Loading tables...")
    df = None
    if (date_ranges is None) or (len(date_ranges) > 0):
        for date_range in date_ranges:
            
            if table_name is not None:
                table = "{}.{}_{}".format(dataset, table_name, date_range)
            else:
                table = "{}.{}_{}".format(dataset, table_type, date_range)

            df_in = gbq.download_table_to_pandas(
                table,
                flush_cache=flush_cache
            )
            if df_in is not None:
                if df is None:
                    df = df_in.copy()
                else:
                    df = df.append(df_in, sort=True, ignore_index=True)
            else:
                if hard_assert:
                    assert df_in is not None, "  Could not load '{}".format(table)
                else:
                    warnings.warn(
                        "  Could not load '{}".format(table),
                        category=UserWarning, 
                        stacklevel=1
                    )
    else:
        table = "{}.{}".format(dataset, table_name)
        df = gbq.download_table_to_pandas(table)

        if df is None:
            if hard_assert:
                assert df is not None, "  Could not load '{}".format(table)
            else:
                warnings.warn(
                    "  Could not load '{}".format(table),
                    category=UserWarning, 
                    stacklevel=1
                )
            return None
    
    do_continue = True
    if check_columns:
        for _,columns in humai_table_field_types[table_type].items():
            for c in columns:
                if c not in df:
                    warnings.warn(
                        f"[load_and_merge_tables] Column {c} is missing...",
                        category=UserWarning, 
                        stacklevel=1
                    )
                    do_continue = False
                    
    if not do_continue:
        return None
    
    print("[load_and_merge_tables] Dropping duplicate rows...")
    print("[load_and_merge_tables]   {:,} rows total.".format(len(df)))
    df.drop_duplicates(inplace=True)
    print("[load_and_merge_tables]   {:,} rows left.".format(len(df)))
    
    print("[load_and_merge_tables] Converting table columns...")
    if datetime_columns is not None:
        columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
    convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)
    
    if table_type == 'transactions' and keep_most_recent_trans_id:
        print("[load_and_merge_tables] Keeping only the most recent Transaction ID, if there's multiples...")
        print("[load_and_merge_tables]   {:,} rows total.".format(len(df)))
        keep_most_recent_transaction_id(df)
        print("[load_and_merge_tables]   {:,} rows left.".format(len(df)))
    
    return df


def upload_and_append_table_unique(
    gbq,
    dataset_table,
    df,
    unique_subset,
    datetime_columns=[],
    table_type=None,
    check_columns=False,
    flush_cache=True,
):
    df_existing = load_table(
        gbq,
        dataset_table,
        table_type=table_type,
        datetime_columns=datetime_columns,
        check_columns=check_columns,
        flush_cache=flush_cache,
    )
    
    if df_existing is not None:
    
        print(f"Size of existing table before appending: {len(df_existing):,}")
        df_existing = df_existing.append(df, sort=True, ignore_index=True)
        df_existing.drop_duplicates(subset=unique_subset, keep='first', inplace=True)
        print(f"Size of table after appending and removing duplicates: {len(df_existing):,}")
    
        gbq.upload_and_replace_table(
            df_existing,
            dataset_table,
        )
    else:
        gbq.upload_and_replace_table(
            df,
            dataset_table,
        )        

def merge_dataframes(dataframes, on, how='left', overwrite_cols=False):
    
    df_out = None
    for df in dataframes:
        if df_out is None:
            df_out = df
        else:
            if not overwrite_cols:
                cols_to_keep = [on] + [c for c in df.columns if c not in df_out]
                df_out = df_out.merge(
                    df[cols_to_keep],
                    on=on,
                    how='left',
                    copy=False,
                )
            else:
                for c in df.columns:
                    if c in df_out:
                        del df_out[c]
                df_out = df_out.merge(
                    df,
                    on=on,
                    how='left',
                    copy=False,
                )
                
    return df_out


def keep_most_recent_transaction_id(df, inplace=True):
    keep_most_recent_id(df, id_column='Transaction_ID', inplace=inplace)
    
    
def keep_most_recent_id(df, id_column='Transaction_ID', inplace=True):
    if not inplace:
        return df.rename_axis('_index').sort_values(by=[id_column,'Datetime','_index']).drop_duplicates(subset=id_column, keep='last')

    df.rename_axis('_index').sort_values(by=[id_column,'Datetime','_index'], inplace=True)
    df.drop_duplicates(subset=id_column, keep='last', inplace=True)
    
    
def force_string(value):
    return 'S' + str(value)#'S' + str(value)


def get_papermill_param(param, type_str='str', vis=None):
    
    if vis is None:
        vis = globals()
        
    var_lookup = {
        'int' : 0,
        'float' : 0.0,
        'list' : [],
        'dict' : {},
        'bool' : False,
        'str' : '',
    }

    value = var_lookup.get(type_str, None)
    if param in vis:
        
        value = vis[param] if vis[param] is not None else var_lookup.get(type_str, None)
        print("PAPERMILL -> Set '{}' to {}".format(
            param, value
        ))
            
    return value


def gbq_friendly_column_name(column):
    
    col_name = column
    if str.isdigit(col_name[0]):
        col_name = '_' + col_name

    #col_name = re.sub(r'[À-ÖØ-öø-ÿ]', '_', col_name)
    return unidecode.unidecode(re.sub(r'[^\w]', '_', col_name).replace('-','_')).replace('+','plus').replace('-','to')


def bq_friendly_column_names(columns):

    return [gbq_friendly_column_name(c) for c in columns]
    

def column_stats(df, col_name, n=-1):
    """
    Print out column statistics.
    """
    
    print("-------------------------")
    print("COLUMN_STATS for '{}'".format(col_name))
    isna = df[col_name].isna().sum()
    print("NaNs: {:,} out of {:,} ({:.2f}%)".format(isna, len(df), 100.0*isna/len(df)))
    print("== Describe ==")
    print(df[col_name].describe())
    
    if n == -1:
        n = df[col_name].nunique()
        
    if not is_float_dtype(df[col_name]):
        print("== Value counts ==")
        print(df[col_name].value_counts()[:n])
        print("\n  Total unique values: {:,}\n".format(df[col_name].nunique()))
        
    print("== Head ==")
    print(df[col_name].head(n=7))
    
    
def column_hist(df, col_name):
    
    fig = px.histogram(df, x=col_name)
    fig.update_layout(
        title="Histogram of '{}'".format(col_name),
        xaxis_title=col_name,
        yaxis_title='Number of customers'
    )
    #fig.show()
    
    return fig
    
          
def columns_as_str(df, columns, uppercase=False, drop_empty=False, replace_na_value=''):
          
    """
    Convert columns to string, and convert to uppercase if specified.

    Note: This is done inplace.
    """
    
    if columns is None:
        columns = df.columns
        
    if isinstance(columns, str):
        columns = [columns]

    if not isinstance(columns, list):
        raise Exception("Expecting columns to be a list.")
    
    for c in columns:
        if c in df:
            if uppercase:    
                # pandas 0.25.0
                #df.loc[:,c] = df[c].astype(str, skipna=True).str.upper()
                # pandas 1.0.0
                df.loc[:,c] = df[c].astype(str).str.upper()
            else:
                # pandas 0.25.0
                #df.loc[:,c] = df[c].astype(str, skipna=True)
                # pandas 1.0.0
                df.loc[:,c] = df[c].astype(str)

            if drop_empty:
                df.loc[:,c].replace(to_replace='', value=np.nan, inplace=True)
                df.dropna(subset=[c], inplace=True)
            else:
                df.loc[:,c].fillna(replace_na_value, inplace=True)

        else:
            print("Column '{}' doesn't exist.".format(c))

            
def columns_as_datetime(df, columns, datetime_format='%Y-%m-%d %H:%M:%S'):
    
    """
    Convert columns to time-zone-agnostic datetime.

    Note: This is done inplace.
    """
    
    if columns is None:
        columns = df.columns
        
    if isinstance(columns, str):
        columns = [columns]

    if not isinstance(columns, list):
        raise Exception("Expecting columns to be a list.")

    for c in columns:
        if c in df:
            df.loc[:,c] = pd.to_datetime(df[c], format=datetime_format, utc=True).dt.tz_convert(None)
        else:
            print("Column '{}' doesn't exist.".format(c))


def columns_as_float(df, columns, na_replace_value=None, drop_na=True):
    
    """
    Convert columns to float.

    Note: This is done inplace.
    """
    
    if columns is None:
        columns = df.columns
        
    if isinstance(columns, str):
        columns = [columns]
        
    if not isinstance(columns, list):
        raise Exception("Expecting columns to be a list.")

    if na_replace_value is None:
        na_replace_value = np.nan
    else:
        drop_na = False
        
    for c in columns:
        if c in df and not is_float_dtype(df[c]):
            
            if is_string_dtype(df[c]):
                df.loc[:,c].replace({
                    '' : np.nan, 'nan' : np.nan, 'NaN' : np.nan, 'None' : np.nan
                }, inplace=True)
                
            if drop_na:
                df.dropna(subset=[c], how='any', inplace=True)
            else:
                df.loc[:,c].fillna(na_replace_value, inplace=True)

            df.loc[:,c] = df[c].astype('float64')
        else:
            if is_float_dtype(df[c]):
                if drop_na:
                    df.dropna(subset=[c], how='any', inplace=True)
                else:
                    df.loc[:,c].fillna(na_replace_value, inplace=True)
                
            if c not in df:
                print("Column '{}' doesn't exist.".format(c))

        

def columns_as_int(df, columns, na_replace_value=None, drop_na=True):
    
    """
    Convert string/float columns to integer.

    Note: This is done inplace.
    """
    
    if columns is None:
        columns = df.columns
        
    if isinstance(columns, str):
        columns = [columns]
        
    if not isinstance(columns, list):
        raise Exception("Expecting columns to be a list.")
        
    if not drop_na and na_replace_value is None:
        na_replace_value = pd.NA

    columns_as_float(df, columns, na_replace_value, drop_na)
    
    for c in columns:
        if c in df:
            df.loc[:,c] = df[c].astype('int64')
        else:
            print("Column '{}' doesn't exist.".format(c))
            
            
def agg_first(series, **kwargs):
    return series.iloc[0]

def agg_last(series, **kwargs):
    return series.iloc[-1]

def agg_nth(series, **kwargs):
    n = kwargs['n']
    if n >= len(series):
        return np.nan
    
    return series.iloc[n]

def agg_list(series, **kwargs):
    return "{}".format(list(series.to_numpy()))


def agg_sequence(series, **kwargs):
    
    d = kwargs['delimiter']
    s = d.join(item for item in series.to_numpy().astype(str))
    
    if kwargs['do_sorting']:
        return d.join(item for item in np.sort( list(set(s.split(d))) ))
    
    return d.join(item for item in list(set(s.split(d))))


def agg_umode(series):
    modes = series.mode()
    return modes.iloc[0]

    
def aggregate_columns(df, agg_type, agg_column, columns, copy_agg_column=False, **kwargs):
    
    """
    Aggregate values of columns, collapsing them.

    Note: This is not done inplace.
    """
    
    if agg_column not in df:
        print("[aggregate_columns] '{}' does not exist!!".format(agg_column))
        return None
    
    if len(columns) == 0:
        return None
    
    for c in columns:
        if c not in df:
            print("[aggregate_columns] '{}' does not exist!!".format(c))
    
    aggregators = {
        'mean' : pd.DataFrame.mean,
        'sum'  : pd.DataFrame.sum,
        'min'  : pd.DataFrame.min,
        'max'  : pd.DataFrame.max,
        'std'  : pd.DataFrame.std,
        'any'  : pd.DataFrame.any,
        'mode' : agg_umode,
        'first': agg_first,
        'last' : agg_last,
        'nth'  : agg_nth,
        'list' : agg_list,
        'seq'  : agg_sequence,
    }
    
    if agg_type in aggregators:
        func = aggregators[agg_type]
    else:
        raise NameError('[aggregate_columns] Need to add {} to transform function'.format(agg_type))
        
    cols = [c for c in columns if c in df]
    assert len(cols) > 0, f"[aggregate_columns] The columns [{columns}] to aggregate {agg_type} over is the empty array, which means we will aggregate over ALL columns. Do you mean to do this?"
    return df.groupby(by=agg_column)[cols].agg(func, **kwargs)




def transactions_cleanup(df, datetime_column, tid_column = 'Transaction_ID', cid_column = 'Customer_ID', datetime_format='%Y-%m-%d %H:%M:%S'):
    
    """
    Perform common transaction data cleanup.

    Note: This is done inplace.
    """
    
    df.loc[:,datetime_column] = pd.to_datetime( df[datetime_column], format=datetime_format)
    df.loc[:,datetime_column].dropna(inplace=True)

    # Get transactions date range
    print('Date range from {0} to {1}'.format(df[datetime_column].min(), df[datetime_column].max())) 
    
    # Remove column 'Unnamed: 0' if it exists
    if 'Unnamed: 0' in df:
        df.drop('Unnamed: 0', axis=1, inplace=True)
        print("Removed unnamed index column")

    print("Before dropping empty Transaction IDs: {:,}".format(len(df)))

    # Drop Transaction_ID's that are empty
    df.loc[:,tid_column].replace('', np.nan, inplace=True)
    df.dropna(subset=[tid_column], inplace=True)

    print("After: {:,}".format(len(df)))
    
    print("Before dropping empty Customer IDs: {:,}".format(len(df)))

    # Drop [cid]'s that are empty
    df.loc[:,cid_column].replace('', np.nan, inplace=True)
    df.dropna(subset=[cid_column], inplace=True)

    print("After: {:,}".format(len(df)))


def email_ensure_zero_or_one_activity_per_broadcast(
    df, cid_column = 'Customer_ID', activity_column = 'Activity', 
    activity_datetime_column = 'Activity_Datetime',
    verbose=True
):

    """
    Ensure each email sent to each customer has only one event: either a delivery, an open, a click, or a transaction,
    in that order. Keep only the first (in time) occurrence of it, in case of duplicate events.

    Note: This is done inplace.
    """
    
    conv_table = {
        'DELIVERY' : 0,
        'OPEN' : 1,
        'UNSUBSCRIBE' : 2,
        'CLICK' : 3,
        'TRANSACTION' : 4
    }
    
    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] converting activities to ordinals...")
    
    df.loc[:,'activity_id'] = df[activity_column].apply(
        lambda x: conv_table.get(x,len(conv_table))
    )
 
    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] get max...")

    result = df.groupby(by=[cid_column, 'Broadcast_ID'])['activity_id'].max()
    
    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] determining which rows to keep...")
        
    df.loc[:,'keep'] = df.apply(lambda x: 1 if result[x[cid_column],x['Broadcast_ID']] == x['activity_id'] else 0, axis=1)

    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] dropping...")
    df.drop(index=df[df.keep == 0].index, inplace=True)
    
    if verbose:
        print("[email_ensure_zero_or_one_activity_per_broadcast] sorting and keeping only the first occurence of the activity...")
        
    df.sort_values(by=[cid_column, 'Broadcast_ID', activity_datetime_column], inplace=True)
    df.drop_duplicates(subset=[cid_column, 'Broadcast_ID'], keep='first', inplace=True)

    del df['keep']
    del df['activity_id']



def email_keep_activities(df, activities):
    
    """
    Keep only certain email activities, remove the rest.

    Note: This is done inplace.
    """
    
    keep = df.Activity.isin(activities)
    df.drop(index=df[~keep].index, inplace=True)
    
    
def email_activity_stats(df, activity_column='Activity'):
    
    n = []
    n.append(len(df[df[activity_column] == 'DELIVERY']))
    n.append(len(df[df[activity_column] == 'OPEN']))
    n.append(len(df[df[activity_column] == 'CLICK']))
    n.append(len(df[df[activity_column] == 'TRANSACTION']))
    n.append(len(df[df[activity_column] == 'UNSUBSCRIBE']))
    n.append(len(df) - np.sum(n))
    N = np.sum(n)
    
    stats = {
        'Delivered' : n[0],
        'Delivered_perc' : 100.0 * n[0]/N,
        'Opens' : n[1],
        'Opens_perc' : 100.0 * n[1]/N,
        'Clicks' : n[2],
        'Clicks_perc' : 100.0 * n[2]/N,
        'Transactions' : n[3],
        'Transactions_perc' : 100.0 * n[3]/N,
        'Unsubscribes' : n[4],
        'Unsubscribes_perc' : 100.0 * n[4]/N,
        'Other' : n[5],
        'Other_perc' : 100.0 * n[5]/N,
        'Total' : np.sum(n)
    }

    return stats


def print_email_activity_stats(df, activity_column='Activity'):
    
    stats = email_activity_stats(df, activity_column=activity_column)
    
    print("Deliveries:   {:,} ({:.2f} %)".format(stats['Delivered'], stats['Delivered_perc']))
    print("Opens:        {:,} ({:.2f} %)".format(stats['Opens'], stats['Opens_perc']))
    print("Clicks:       {:,} ({:.2f} %)".format(stats['Clicks'], stats['Clicks_perc']))
    print("Transactions: {:,} ({:.2f} %)".format(stats['Transactions'], stats['Transactions_perc']))
    print("Unsubscribes: {:,} ({:.2f} %)".format(stats['Unsubscribes'], stats['Unsubscribes_perc']))
    print("Other:        {:,} ({:.2f} %)".format(stats['Other'], stats['Other_perc']))
    print("Total:        {:,}".format(stats['Total']))
    
    
def transaction_stats(
    df,
    cid_column='Customer_ID',
    datetime_column='Datetime',
    max_bins=100
):
    
    df.loc[:,'rank'] = df.sort_values(by=[cid_column, datetime_column]). \
        groupby(by=[cid_column])[datetime_column].rank(method='dense')
    #df.loc[:,'trans'] = 1
    
    n_trans = df.sort_values(by=[cid_column, datetime_column]).groupby(by=[cid_column])['rank'].max()
    #n_trans = df.sort_values(by=[cid_column, datetime_column]).groupby(by=[cid_column])['trans'].sum()

    df_n_trans = pd.DataFrame()
    df_n_trans.loc[:,cid_column] = n_trans.index
    df_n_trans.loc[:,'n_trans'] = n_trans.to_numpy()

    hist_n_trans = df_n_trans['n_trans'].value_counts()

    df_trans_hist = pd.DataFrame()
    df_trans_hist['n_trans'] = hist_n_trans.index
    df_trans_hist['count'] = hist_n_trans.values
    df_trans_hist.sort_values(by=['n_trans'], inplace=True)
    print(df_trans_hist.head(max_bins))
    
    fig = column_hist(df_n_trans, 'n_trans')
    
    del df['rank']
    
    print("Period {} to {}".format(df[datetime_column].min(), df[datetime_column].max()))
    print("Total sales: {:,} ({:,} customers)".format(len(df), df[cid_column].nunique()))
    print("Order segment 3+: {:,} customers".format(df_trans_hist[df_trans_hist['n_trans'] > 2]['count'].sum()))
    
    return fig, df_trans_hist


def keep_subset_customers(dataframes, cid_column, k, seed=1):
    """
    Keep a subset of customers among all dataframes.
    
    Input:
    dataframes - Either a single dataframe or a list of dataframes.
    
    Is not done inplace.
    """
    import random
    
    random.seed(seed)
    
    dataframes_in = dataframes
    if isinstance(dataframes, list):
        df = dataframes[0]
    else:
        df = dataframes
        dataframes_in = [dataframes]
    
    print("{:,} unique customers".format(df[cid_column].nunique()))
    
    cids = list(df[cid_column].unique())
    to_keep_cids = random.sample(cids, k=k)
    to_keep_cids_map = {cid:1 for cid in to_keep_cids}

    dataframes_out = []
    for df in dataframes_in:
        df.loc[:,'keep'] = df[cid_column].apply(lambda x: to_keep_cids_map.get(x,0))
        dataframes_out.append( df.drop(index=df[df.keep == 0].index))
    
    print("{:,} unique customers after sampling".format(dataframes_out[0][cid_column].nunique()))
    
    if isinstance(dataframes, list):
        return dataframes_out
    
    return dataframes_out[0]


def filter_dataframe(df, match_filters=None, contains_filters=None):
    #
    # Perform any filtering
    #
    if match_filters is not None:
        for c,v in match_filters.items():
            if c in df:
                filter_vals = '|'.join(v)

                column_stats(df, c)
                print(f"Before 'match' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
                df = df.loc[df[c].str.match(filter_vals),:].copy()
                print(f"After 'match' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
                column_stats(df, c)
                
    if contains_filters is not None:
        for c,v in contains_filters.items():
            if c in df:
                filter_vals = '|'.join(v)

                column_stats(df, c)
                print(f"Before 'contains' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
                df = df.loc[df[c].str.contains(filter_vals),:].copy()
                print(f"After 'contains' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
                column_stats(df, c)

    return df


def mann_whitney_test(dfA, dfB, column, group_A_name='grp_A', group_B_name='grp_B', scale = 1.0, unit='', verbose=False):
    
    from scipy.stats import mannwhitneyu as mw
    
    try:
        dfa = dfA[~dfA[column].isnull()]
        dfb = dfB[~dfB[column].isnull()]
        
        p = mw(dfa[column], dfb[column])[1]
        
        na = len(dfa)
        nb = len(dfb)
        
        if verbose:
            print('{}: sum/mean (scale={})/median (scale={}) \n A={:,}/{:.3f}{}/{:.3f}{} (n={:,}), B={:,}/{:.3f}{}/{:.3f}{} (n={:,}), p={:.3f}'.format(
                column, scale, scale,
                int(dfa[column].sum()),
                dfa[column].mean() * scale,
                unit,
                dfa[column].median() * scale,
                unit,
                na, 
                int(dfb[column].sum()),
                dfb[column].mean() * scale, 
                unit,
                dfb[column].median() * scale,
                unit,
                nb, 
                p
            ))
        
        data = []
        data.append((
            column,
            
            #int(dfa[column].sum()),
            dfa[column].mean() * scale,
            dfa[column].median() * scale,
            na,
            
            #int(dfb[column].sum()),
            dfb[column].mean() * scale,
            dfb[column].median() * scale,
            nb,
            
            p
        ))
        
        #df_out = pd.DataFrame(data, columns=[
        #    f'driver',
        #    #f'sum_{group_A_name}',
        #    f'mean_{group_A_name}',
        #    f'median_{group_A_name}',
        #    f'size_{group_A_name}',
        #    
        #    #f'sum_{group_B_name}',
        #    f'mean_{group_B_name}',
        #    f'median_{group_B_name}',
        #    f'size_{group_B_name}',
        #    
        #    f'mann_whitney_p_value',
        #])
        df_out = pd.DataFrame(data, columns=[
            'column',
            #f'sum_{group_A_name}',
            '{}_mean'.format(group_A_name),
            '{}_median'.format(group_A_name),
            '{}_size'.format(group_A_name),
            
            #f'sum_{group_B_name}',
            '{}_mean'.format(group_B_name),
            '{}_median'.format(group_B_name),
            '{}_size'.format(group_B_name),
            
            'mann_whitney_p_value',
        ])
        
        #print(df_out.head())
        
        return df_out, p
            
    except:
        
        return None, np.nan
    
    return None, np.nan


def parallelise_dataframe(df, func, on_cids=False, n_cores=2):
    if on_cids:
        print("[parallelise_data_frame_on_cids] Assigning Customer_IDs to cores...")
        cid2batch = {cid:(i % n_cores) for i,cid in enumerate(df.Customer_ID.unique())}
        df.loc[:,'core'] = df.Customer_ID.apply(lambda cid: cid2batch.get(cid, 0))
        df_split = [df[df.core == i] for i in range(n_cores)]
    else:
        df_split = np.array_split(df, n_cores)

    print(f"Executing on {n_cores} cores...")
    pool = Pool(n_cores)
    df = pd.concat(pool.map(func, df_split))

    pool.close()
    pool.join()

    return df



# Bayesian weekday probability model

import pandas as pd
import numpy as np

import random

def histogr(df):
    vc = df.value_counts().sort_index()
    return vc# / vc.sum()

# Calculate likelihood:
# Lambda(D) := P(H | D) = Probability that the most popular day H:=argmax(H), given the
# most recent day D.

def CalcLikelihoodOverallMax(dfg_customers):
    ii = 0
    N = len(dfg_customers)
    likelihood = [np.zeros(7) for x in range(0,7)]
    cids = []
    hs = []
    for cid, g in dfg_customers:
        h = histogr(g['transaction_weekday'])
        hist = np.zeros(7)
        hist[h.index] = h.values

        mr_wday = int( g['transaction_weekday'].iloc[-1] )

        hist[mr_wday] = max(0, hist[mr_wday] - 1) # Subtract most recent transaction
        #print(hist)
        most_popular = np.argwhere(hist == np.amax(hist))
        most_popular = most_popular.flatten().tolist()
        #print(most_popular)
        most_popular = int(random.choice(most_popular))

        #print('mr: {0}, pop: {1}'.format(mr_wday, most_popular))

        likelihood[mr_wday][most_popular] = 1 + likelihood[mr_wday][most_popular]

        ii = 1 + ii
        
        if ii % 1000 == 0:
            print('{0}/{1}'.format(ii, N))
            
        cids.append(cid)
        hs.append(hist)

    for i in range(0,len(likelihood)):
        ss = np.sum( likelihood[i] )
        if ss > 0.0:
            likelihood[i] = likelihood[i] / ss
            
    return likelihood, cids, hs

# Calculate likelihood:
# Lambda(D) := P(H | D) = Probability that the most popular day of the last m days,
# H := argmax(H[-m:]), given the most recent day D.

def CalcLikelihoodRecentMax(dfg_customers, m = 5):
    ii = 0
    N = len(dfg_customers)
    likelihood = [np.zeros(7) for x in range(0,7)]
    cids = []
    hs = []
    for cid, g in dfg_customers:
        h = histogr(g['transaction_weekday'].iloc[-m:])
        hist = np.zeros(7)
        hist[h.index] = h.values
        #print('Hist: {0}'.format(hist))
        
        mr_wday = int( g['transaction_weekday'].iloc[-1] )
        hist[mr_wday] = max(0, hist[mr_wday] - 1) # Subtract most recent transaction
        
        #print('Most recent m: {0}'.format(g['transaction_weekday'].iloc[-m:].values))
        #print('All: {0}'.format(g['transaction_weekday'].values))
        most_popular = np.argwhere(hist == np.amax(hist))
        most_popular = most_popular.flatten().tolist()
        #print(most_popular)
        most_popular = int(random.choice(most_popular))

        #print('mr: {0}, pop: {1}'.format(mr_wday, most_popular))

        likelihood[mr_wday][most_popular] = 1 + likelihood[mr_wday][most_popular]

        ii = 1 + ii
        
        if ii % 1000 == 0:
            print('{0}/{1}'.format(ii, N))
            
        cids.append(cid)
        hs.append(hist)

    for i in range(0,len(likelihood)):
        ss = np.sum( likelihood[i] )
        if ss > 0.0:
            likelihood[i] = likelihood[i] / ss
            
    return likelihood, cids, hs

# Calculate histories only:

def CalcHistories(dfg_customers, m = 5):
    ii = 0
    N = len(dfg_customers)
    cids = []
    hs = []
    for cid, g in dfg_customers:
        h = histogr(g['transaction_weekday'].iloc[-m:])
        hist = np.zeros(7)
        hist[h.index] = h.values
        
        # Get most recent.
        mr_wday = int( g['transaction_weekday'].iloc[-1] )
        hist[mr_wday] = max(0, hist[mr_wday] - 1) # Subtract most recent transaction
        
        ii = 1 + ii
        if ii % 1000 == 0:
            print('{0}/{1}'.format(ii, N))
            
        cids.append(cid)
        hs.append(hist)
            
    return cids, hs

def BayesianProbFromHist(prior, likelihood, hist=[]):
    
    if len(hist) == 0:
        return prior

    n = len(prior)
    p_d_h = [0] * 7
    
    most_popular = np.argwhere(hist == np.amax(hist))
    most_popular = most_popular.flatten().tolist()
    #print(most_popular)
    most_popular = int(random.choice(most_popular))
    
    for d in range(0,7):
        p_d_h[d] = likelihood[d][most_popular] * prior[d]

    return p_d_h 
    
    
    

    


import inspect
import dill as pickle
from pathlib import Path

import joblib

from humailib.cloud_tools import GoogleCloudStorage

class FunctionsGherkin:
    """
    This class finds, saves, loads and reinstates all local functions that have a certain
    parameter signature.
    """
    
    def __init__(self):
        
        self.funcs = []
        return

    def find(self, vars_in_scope = None, param_signature = ['X','params']):
        
        if vars_in_scope is None:
            vars_in_scope = globals().copy()

        self.funcs = []
        for var_name in vars_in_scope:
            if inspect.isfunction( vars_in_scope[var_name] ):
                params = list( inspect.signature(vars_in_scope[var_name]).parameters )
                if params == param_signature:
                    self.funcs.append( (var_name, vars_in_scope[var_name], params) )
                    
                    
    def reinstate(self, vars_in_scope = None):
        
        if vars_in_scope is None:
            vars_in_scope = globals()
            
        for fname, fobj, params in self.funcs:
            print("[FunctionsGherkin::reinstate] Adding function '{}', with parameters = {}".format(fname, params))
            vars_in_scope[fname] = fobj
            
        return True
    

    def save(self, filename, gcs=None, cache_dir='./cache'):
        
        if 'gs://' in filename:
            if gcs is None:
                gcs = GoogleCloudStorage()
                
            cache_filename = cache_dir + '/' + filename.split('/')[-1]
            file = open(cache_filename, 'wb')
            pickle.dump(self.funcs, file)
            file.close()
            
            gcs.upload_file(cache_filename, filename)
        else:
            file = open(filename, 'wb')
            pickle.dump(self.funcs, file)
            file.close()
            
        return True
    
    
    def load(self, filename, gcs=None, cache_dir='./cache'):
        
        if 'gs://' in filename:
            if gcs is None:
                gcs = GoogleCloudStorage()
                
            cache_filename = Path(cache_dir + '/' + filename.split('/')[-1])
            gcs.download_file(filename, cache_filename)
            cache_filename = cache_dir + '/' + filename.split('/')[-1]
            file = open(cache_filename, 'rb')
            funcs = pickle.load(file)
            file.close()  

            self.funcs = funcs
        else:
            file = open(filename, 'rb')
            funcs = pickle.load(file)
            self.funcs = funcs
            file.close()  
            
        return True
    
    
    def load_and_reinstate(self, filename, vars_in_scope = None, gcs = None, cache_dir='./cache'):
        
        if self.load(filename, gcs=gcs, cache_dir=cache_dir):
            return self.reinstate(vars_in_scope = vars_in_scope)
    
        return False


import pandas as pd
import numpy as np
import datetime

from sklearn.base import BaseEstimator, TransformerMixin


def get_time_of_day(timestamp, morning=12, evening=17):
    """
    """    
    if timestamp.hour < morning:
        return 'morning'
    if timestamp.hour > evening:
        return 'evening'
    return 'daytime'



def get_time_of_day_id(timestamp, morning=12, evening=17):
    """
    """
    if timestamp.hour < morning:
        return 0
    if timestamp.hour > evening:
        return 2
    return 1



def get_training_features_and_labels_tod(
    df, activity, 
    max_hist_len, cid_column, 
    activity_column, datetime_column, verbose=False
):
    """
    """
    
    print("Calculating time of day columns...")
    df_tod = pd.get_dummies( df[datetime_column].apply(lambda x: get_time_of_day(x)) )
    df = df.join( df_tod )

    print("Calculating groups...")
    df.loc[:,'observed'] = df[activity_column].apply(lambda x: 1 if x == activity else 0)
    
    dfg_customers = df.sort_values(by=[cid_column, datetime_column]).groupby(by=[cid_column])
    nc = len(dfg_customers)
    print("Number of customers: {:,} ({:,} rows)".format(nc, len(df)))
     
    features = []

    ii = 0
    tod = [0,0,0]

    for cid, g in dfg_customers:

        if ii % 5000 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))

        if len(g) > 1:

            ep = np.random.randint(1, len(g)-1) if len(g) > 2 else 1
            data = g.iloc[ max(ep-max_hist_len, 0):ep+1, ]
            
            ob_row = data.iloc[-1,:]
            
            observed = ob_row['observed']
            
            #tod[0] = data['morning'].sum()
            #tod[1] = data['daytime'].sum()
            #tod[2] = data['evening'].sum()
            tod[0] = data.apply(lambda x: x['morning']*x['observed'], axis=1).sum()
            tod[1] = data.apply(lambda x: x['daytime']*x['observed'], axis=1).sum()
            tod[2] = data.apply(lambda x: x['evening']*x['observed'], axis=1).sum()
            tod[0] = tod[0] - ob_row['morning']*ob_row['observed']
            tod[1] = tod[1] - ob_row['daytime']*ob_row['observed']
            tod[2] = tod[2] - ob_row['evening']*ob_row['observed']

            n_act = data['observed'].sum() - observed
            n_nonact = (len(data) - 1) - n_act
            last_email_date = data.iloc[-2,:][datetime_column]
            delta_t = data.iloc[-1,:][datetime_column] - last_email_date

            if verbose:
                print(cid)
                print(data)
                print('observed {}, delta_t {}, n_act {}, n_nonact {}, m {}, d {}, e {}, tod {}, ldate {}'.format(
                    observed, delta_t, n_act, n_nonact, 
                    tod[0], tod[1], tod[2], email_tod_id, last_email_date))

            features.append((cid, observed, delta_t, n_act, n_nonact, 
                             tod[0], tod[1], tod[2], 
                             last_email_date))

        ii = 1 + ii

    print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))
    
    df_features = pd.DataFrame(features, columns=[cid_column,'observed','delta_t', 
                                                  'n_act', 'n_nonact', 
                                                  'morning', 'daytime', 'evening', 
                                                  'last_email_date'])
    df_features['delta_t'] = df_features['delta_t'].apply(
        lambda x: max(0, x.days + x.seconds/(24*60*60)))
    df_features['total_num_emails'] = df_features.apply(
        lambda x: x.n_act+x.n_nonact, axis=1)

    return df_features


"""
"""
class EmailTrainingFeaturesTODEstimator(BaseEstimator, TransformerMixin):
    
    def __init__(
        self, activity, max_hist_len, cid_column='Customer_ID', activity_column='Activity', datetime_column='Activity_Datetime'
    ):
        self.activity = activity
        self.max_hist_len = max_hist_len
        self.cid_column = cid_column
        self.activity_column = activity_column
        self.datetime_column = datetime_column
        
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[EmailTrainingFeaturesTODEstimator] transform {:,}...".format(len(X)))
        print(X[[self.cid_column, self.datetime_column, self.activity_column]].head())

        df = get_training_features_and_labels_tod(
            X,
            activity = self.activity,
            max_hist_len = self.max_hist_len,
            cid_column = self.cid_column,
            activity_column = self.activity_column,
            datetime_column = self.datetime_column
        )
        
        return df
    

"""
"""
def get_prediction_features_tod(
    df, activity, max_hist_len, 
    cid_column, activity_column, datetime_column, verbose=False
):
    
    print("Calculating time of day columns...")
    df_tod = pd.get_dummies( df[datetime_column].apply(lambda x: get_time_of_day(x)) )
    df = df.join( df_tod )

    print("Calculating groups...")
    df.loc[:,'observed'] = df[activity_column].apply(lambda x: 1 if x == activity else 0)
    dfg_customers = df.sort_values(by=[cid_column, datetime_column]).groupby(by=[cid_column])
    nc = len(dfg_customers)
    print("Number of customers: {:,} ({:,} rows)".format(nc, len(df)))
  
    features = []

    ii = 0
    tod = [0,0,0]

    for cid, g in dfg_customers:

        if ii % 5000 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))

        if len(g) >= 1:
            data = g.iloc[-max_hist_len:]
            #print(data)

            #tod[0] = data['morning'].sum()
            #tod[1] = data['daytime'].sum()
            #tod[2] = data['evening'].sum()
            tod[0] = data.apply(lambda x: x['morning']*x['observed'], axis=1).sum()
            tod[1] = data.apply(lambda x: x['daytime']*x['observed'], axis=1).sum()
            tod[2] = data.apply(lambda x: x['evening']*x['observed'], axis=1).sum()

            n_act = data['observed'].sum()
            n_nonact = len(data) - n_act
            last_email_date = data.iloc[-1,:][datetime_column]
           
            if verbose:
                print(data)
                print('n_act {}, n_nonact {}, m {}, d {}, e {}, ldate {}'.format(
                    n_act, n_nonact, tod[0], tod[1], tod[2], last_email_date))

            features.append((cid, n_act, n_nonact, 
                             tod[0], tod[1], tod[2], 
                             last_email_date))

        ii = 1 + ii

    print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))
    
    df_features = pd.DataFrame(features, columns=[
        cid_column, 'n_act', 'n_nonact', 'morning', 'daytime', 'evening', 'last_email_date'
    ])
    df_features['total_num_emails'] = df_features.apply(lambda x: x.n_act+x.n_nonact, axis=1)

    return df_features


"""
"""
class EmailPredictionFeaturesTODEstimator(BaseEstimator, TransformerMixin):
    
    def __init__(
        self, activity, max_hist_len, cid_column, activity_column, datetime_column
    ):
        self.activity = activity
        self.max_hist_len = max_hist_len
        self.cid_column = cid_column
        self.activity_column = activity_column
        self.datetime_column = datetime_column
        
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[EmailPredictionFeaturesTODEstimator] transform {:,}...".format(len(X)))

        df = get_prediction_features_tod(
            X,
            activity = self.activity,
            max_hist_len = self.max_hist_len,
            cid_column = self.cid_column,
            activity_column = self.activity_column,
            datetime_column = self.datetime_column
        )
        
        return df
    
    

"""
"""
def get_training_features_basic(
    df, activity, max_hist_len = 16, 
    cid_column='External_ID', activity_column='Activity', datetime_column='Activity_Datetime'
):
    
    df['observed'] = df[activity_column].apply(lambda x: 1 if x == activity else 0)
    dfg_customers = df.sort_values(by=[cid_column, datetime_column]).groupby(by=[cid_column])
    nc = len(dfg_customers)
    print("Number of customers: {:,}".format(nc))
    
    features = []

    ii = 0

    for cid, g in dfg_customers:

        if ii % 5000 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))

        if len(g) > 1:
            #print(cid)
            ep = np.random.randint(1, len(g)-1) if len(g) > 2 else 1
            data = g.iloc[ max(ep-(max_hist_len+1), 0):ep+1, ]
            #print(data)

            observed = data.iloc[-1,:]['observed']

            n_act = data['observed'].sum() - observed
            n_nonact = (len(data) - 1) - n_act
            last_email_date = data.iloc[-1,:][datetime_column]
            delta_t = last_email_date - data.iloc[-2,:][datetime_column] 

            #print('observed {}, delta_t {}, n_act {}, n_nonact {}, ldate {}'.format(observed, delta_t, n_act, n_nonact, last_email_date))

            features.append((cid, observed, delta_t, n_act, n_nonact, last_email_date))

        ii = 1 + ii

    print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))
    
    df_features = pd.DataFrame(features, columns=[cid_column,'observed','delta_t', 'n_act', 'n_nonact', 'last_email_date'])
    df_features['delta_t'] = df_features['delta_t'].apply(lambda x: max(0, x.days + x.seconds/(24*60*60)))
    df_features['total_num_emails'] = df_features.apply(lambda x: x.n_act+x.n_nonact, axis=1)

    return df_features


"""
"""
def get_prediction_features_basic(
    df, activity, timepoint, max_hist_len=16, 
    cid_column='External_ID', activity_column='Activity', datetime_column='Activity_Datetime'
):
    
    df['observed'] = df[activity_column].apply(lambda x: 1 if x == activity else 0)
    dfg_customers = df.sort_values(by=[cid_column, datetime_column]).groupby(by=[cid_column])
    nc = len(dfg_customers)
    print("Number of customers: {:,}".format(nc))
    
    features = []

    ii = 0

    for cid, g in dfg_customers:

        if ii % 5000 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))

        if len(g) >= 1:
            data = g.iloc[-max_hist_len:]
            #print(data)

            observed = data.iloc[-1,:]['observed']

            n_act = data['observed'].sum()
            n_nonact = len(data) - n_act
            last_email_date = data.iloc[-1,:][datetime_column]
            delta_t = timepoint - last_email_date 

            #print('observed {}, delta_t {}, n_act {}, n_nonact {}, ldate {}'.format(observed, delta_t, n_act, n_nonact, last_email_date))

            features.append((cid, observed, delta_t, n_act, n_nonact, last_email_date))

        ii = 1 + ii

    print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))
    
    df_features = pd.DataFrame(features, columns=[cid_column,'observed','delta_t', 'n_act', 'n_nonact', 'last_email_date'])
    df_features['delta_t'] = df_features['delta_t'].apply(lambda x: max(0, x.days + x.seconds/(24*60*60)))
    df_features['total_num_emails'] = df_features.apply(lambda x: x.n_act+x.n_nonact, axis=1)

    return df_features


"""
For a given time period, starting at [start_date] and lasting [timespan_wks] weeks:
Generate [num_emails] predictions for each customer, by taking the [num_emails] days
on which the model predicts the highest probability of activity.

max_slot_space - maximum number of predictions in each timeslot
disable_days - days for which never to generate a prediction
"""
def get_timeslots(
    df, 
    rm, 
    cid_column, covariate_columns, pred_threshold,
    start_date = None, 
    timespan_wks = 1, 
    num_emails = 1, 
    max_slot_space=1000000, disable_days = None
):
    
    days_in_timespan = 7*timespan_wks
    num_timeslots = 3*days_in_timespan
    
    frame = []
    timeslot_hist = [0] * num_timeslots
        
    if disable_days is not None:
        for dis in disable_days:
            timeslot_hist[dis*3:dis*3+3] = [max_slot_space] * 3

    timeslot_data = []

    nn = len(df)
    j = 0
    for _,row in df.iterrows():

        if j % 500 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(j, nn, 100.0*j/nn))
            
        if start_date is None:
            _start_date = row.last_email_date + datetime.timedelta(days=1)
        elif type(start_date) == str:
            _start_date = row[start_date]
        else:
            _start_date = start_date
            
        if type(num_emails) == str:
            _num_emails = row[num_emails]
        else:
            _num_emails = num_emails
        
        day_offset = int( (_start_date - row.last_email_date).days )
        if day_offset < 0:
            print("{} predicting in the past. start date is {}, but last email event date is: {}".format(_start_date, row.last_email_date))
            day_offset = 0

        frame = []
        for day in range(0,days_in_timespan):
            frame.append((day+day_offset, row.n_act, row.n_nonact, row.morning+1, row.daytime  , row.evening  , 0))
            frame.append((day+day_offset, row.n_act, row.n_nonact, row.morning  , row.daytime+1, row.evening  , 1))
            frame.append((day+day_offset, row.n_act, row.n_nonact, row.morning  , row.daytime  , row.evening+1, 2))

        df_to_pred = pd.DataFrame(frame, columns=['delta_t', 'n_act', 'n_nonact', 'morning', 'daytime', 'evening', 'tod'])

        rm.predict(df_to_pred, covariate_columns, pred_threshold=pred_threshold)
        #print(df_to_pred.head(21))
        
        rankings = df_to_pred.sort_values(by=['pred_ranking'], ascending=False)
        #print(rankings)
        for eid in range(_num_emails):
            best_day = int(rankings.iloc[eid,:].delta_t - day_offset)
            best_tod = int(rankings.iloc[eid,:].tod)
            best_slot = best_day*3 + best_tod
            propensity = rankings.iloc[eid,:].pred_ranking
            i = eid
            while timeslot_hist[best_slot] >= max_slot_space and i < len(df_to_pred):
                best_day = int(rankings.iloc[i,:].delta_t - day_offset)
                best_tod = int(rankings.iloc[i,:].tod)
                propensity = rankings.iloc[i,:].pred_ranking
                best_slot = best_day*3 + best_tod  
                i = 1 + i
            if i == len(df_to_pred):
                print("Timeslots full ({:,}/{:,})".format(j, nn))
                best_day = int(rankings.iloc[eid,:].delta_t - day_offset)
                best_tod = int(rankings.iloc[eid,:].tod)
                propensity = rankings.iloc[eid,:].pred_ranking
               
            best_slot = best_day*3 + best_tod
            #print(best_slot)
            timeslot_hist[best_slot] = 1 + timeslot_hist[best_slot]
            best_date = _start_date + datetime.timedelta(days=best_day)
            timeslot_data.append((
                row[cid_column], 
                eid, 
                best_date.strftime('%Y-%m-%d'), 
                best_date.weekday(), 
                best_tod,
                propensity
            ))
            #print("Best slot for {}: (day {}, tod {})".format(
            #    row[cid_column],  best_date.weekday(), best_tod
            #))
            
        j = 1 + j

    print("  {:,}/{:,} ({:.2f} %)".format(j, nn, 100.0*j/nn))
    print("Done.")
    
    #print(timeslot_data)

    df_out = pd.DataFrame(timeslot_data, columns=[
        cid_column, 'Email_ID', 'Date', 'Day_Of_Week', 'Time_Of_Day', 'Propensity'
    ])
    df_out.loc[:,'Date'] = pd.to_datetime(df_out['Date'], format='%Y-%m-%d')
    
    return df_out

"""
For a given time period, starting at [start_date]:
Generate [num_emails] predictions for each customer, by taking the [num_emails] days
on which the model predicts the highest probability of activity, and then spacing
them consecutive weeks apart.

max_slot_space - maximum number of predictions in each timeslot
disable_days - days for which never to generate a prediction
"""
def get_timeslots_consecutive_weeks(
    df, 
    rm, 
    cid_column, covariate_columns, pred_threshold,
    start_date = None, # Can be None, a datetime object, or a string denoting a column
    num_emails = 1, # Can be an integer, or a string denoting a column
    max_slot_space=1000000, disable_days = None
):
    df_out = get_timeslots(
        df,
        rm,
        cid_column, covariate_columns, pred_threshold,
        start_date = start_date,
        timespan_wks = 1,
        num_emails = num_emails,
        max_slot_space = max_slot_space,
        disable_days = disable_days
    )
    
    # We can use 'Email_ID' as the week identifier.
    df_out.loc[:,'Date'] = df_out.apply(lambda x: x.Date+datetime.timedelta(days=7*x.Email_ID), axis=1)
    
    return df_out


from abc import abstractmethod
import requests
import rauth
from rauth import OAuth1Service
import json
import datetime
import sys

import numpy as np
import pandas as pd

import humailib.utils as hu

class MagentoREST:
    
    def __init__(self, access_token, base_url):
        
        self.access_token = access_token
        #self.base_url = 'https://24fashion.nl/rest/default/V1/'
        self.base_url = base_url
                 
    def _get_uri(self, endpoint):
        
        return self.base_url + endpoint
  
    def post(self, endpoint, params, verbose=False):

        url = self._get_uri(endpoint)
        response = requests.post(url, params=params, headers={'Authorization':'Bearer {}'.format(self.access_token)})
        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {}".format(response.status_code))
            print("Response: {}".format(response.text))

        return response
    
    def put(self, endpoint, params, verbose=False):
        
        url = self._get_uri(endpoint)
        response = requests.put(url, params=params, headers={'Authorization':'Bearer {}'.format(self.access_token)})
        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {}".format(response.status_code))
            print("Response: {}".format(response.text))

        return response
    
    def get(self, endpoint, params=None, verbose=False):
        
        url = self._get_uri(endpoint)
        response = requests.get(url, params=params, headers={'Authorization':'Bearer {}'.format(self.access_token)})
        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {}".format(response.status_code))
            print("Response: {}".format(response.text))

        return response, response.json()
    
class MagentoOrdersExtract:
    
    def __init__(self, access_token, base_url):
        
        self.api = MagentoREST(access_token, base_url)
        
    def _extract_field(self, order, field_name):
        
        return order.get(field_name,'') # order[field_name] if field_name in order else ''
        
    def _extract_float(self, order, field_name):
        
        return order.get(field_name,np.nan)#order[field_name] if field_name in order else np.nan

    def _extract_array(self, customer, field_name):
        
        return customer.get(field_name,[''])#customer[field_name] if field_name in customer else ['']
    
    def extract(self, from_date, to_date, page_size = 1000, verbose=False):
        
        """
        Extract all orders between the dates [from_date, to_date)
        """
        
        order_rows = []
        item_rows = []
        
        try:
            params = {
                'searchCriteria[currentPage]':str(1), 
                'searchCriteria[pageSize]':str(1),
                'searchCriteria[filter_groups][0][filters][0][field]':'created_at',
                'searchCriteria[filter_groups][0][filters][0][value]':from_date.strftime('%Y-%m-%d %H:%M:%S'),
                'searchCriteria[filter_groups][0][filters][0][condition_type]':'gteq',
                'searchCriteria[filter_groups][1][filters][0][field]':'created_at',
                'searchCriteria[filter_groups][1][filters][0][value]':to_date.strftime('%Y-%m-%d %H:%M:%S'),
                'searchCriteria[filter_groups][1][filters][0][condition_type]':'lt'
            }
            response, orders = self.api.get('orders', params=params, verbose=False)
            to_id = orders['total_count']
            
            if verbose:
                print("Total number of orders {:,}".format(orders['total_count']))
        except:
            print("Unexpected error: " + sys.exc_info()[0])
        
        order_id = 1
        while order_id <= to_id:
                
            try:
                
                page = int(np.floor((order_id-1)/float(page_size))) + 1
                if verbose:
                    print("Requesting page {:,} of (orders from {:,} to {:,}, total={:,})".format(page, (page-1)*page_size, (page-1)*page_size + page_size, to_id))
                    
                params = {
                    'searchCriteria[currentPage]':str(page), 
                    'searchCriteria[pageSize]':str(page_size),
                    'searchCriteria[filter_groups][0][filters][0][field]':'created_at',
                    'searchCriteria[filter_groups][0][filters][0][value]':from_date.strftime('%Y-%m-%d %H:%M:%S'),
                    'searchCriteria[filter_groups][0][filters][0][condition_type]':'gteq',
                    'searchCriteria[filter_groups][1][filters][0][field]':'created_at',
                    'searchCriteria[filter_groups][1][filters][0][value]':to_date.strftime('%Y-%m-%d %H:%M:%S'),
                    'searchCriteria[filter_groups][1][filters][0][condition_type]':'lt'
                }
                order_id = page_size + order_id
                response, orders = self.api.get('orders', params=params, verbose=False)

                if (200 <= response.status_code <= 299) and 'items' in orders:
                  
                    for order in orders['items']:
                        
                        if 'items' in order and 'entity_id' in order and 'billing_address' in order:
                        
                            items = order['items']
                            billing = order['billing_address']
                            shipping = None
                            shipping_city = ''
                            shipping_country_id = ''
                            shipping_zip = ''
                            shipping_address = ''
                            if 'extension_attributes' in order:
                                if 'shipping_assignments' in order['extension_attributes']:
                                    shipping_data = order['extension_attributes']['shipping_assignments']
                                    if len(shipping_data) > 0:
                                        shipping_data = shipping_data[0]
                                        if 'shipping' in shipping_data:#
                                            if 'address' in shipping_data['shipping']:
                                                shipping = shipping_data['shipping']['address']
                            
                            entity_id = int( self._extract_field(order, 'entity_id') )
                            transaction_id = 'T'+str(entity_id)#hu.force_string(entity_id) # Force it to be a string

                            if True:
                                
                                #if verbose:
                                #    print("  Extracting order {}".format(transaction_id))

                                email = self._extract_field(order, 'customer_email')
                                firstname = self._extract_field(order, 'customer_firstname')
                                lastname = self._extract_field(order, 'customer_lastname')

                                billing_city = self._extract_field(billing, 'city')
                                billing_country_id = self._extract_field(billing, 'country_id')
                                billing_zip = self._extract_field(billing, 'postcode')
                                billing_address = ' '.join(self._extract_array(billing, 'street'))
                                if shipping is not None:
                                    shipping_city = self._extract_field(shipping, 'city')
                                    shipping_country_id = self._extract_field(shipping, 'country_id')
                                    shipping_zip = self._extract_field(shipping, 'postcode')
                                    shipping_address = ' '.join(self._extract_array(shipping, 'street'))

                                order_rows.append(tuple([
                                    transaction_id, 
                                    'I'+self._extract_field(order,'increment_id'),
                                    self._extract_float(order,'base_currency_code'),
                                    self._extract_float(order,'base_grand_total'), 
                                    self._extract_float(order,'discount_amount'), 
                                    self._extract_float(order,'subtotal'), 
                                    self._extract_float(order,'tax_amount'), 
                                    self._extract_float(order,'total_paid'),
                                    self._extract_float(order,'total_item_count'),
                                    self._extract_field(order,'customer_is_guest'),
                                    email, 
                                    self._extract_field(billing,'entity_id'), 
                                    firstname, lastname,
                                    billing_address,
                                    billing_zip,
                                    billing_city, billing_country_id, 
                                    shipping_address,
                                    shipping_zip,
                                    shipping_city, shipping_country_id, 
                                    self._extract_field(order,'created_at'), 
                                    self._extract_field(order,'status'),
                                    self._extract_field(order,'store_name'), 
                                    self._extract_field(order,'store_id'),
                                    'product',
                                    False,
                                    0.0,
                                    '',
                                ]))

                                for item in items:
                                    item_rows.append(tuple([
                                        transaction_id,
                                        self._extract_field(item,'created_at'),
                                        self._extract_field(item,'name'), 
                                        self._extract_field(item,'sku'), 
                                        self._extract_field(item,'product_id'),
                                        self._extract_float(item,'discount_amount'),
                                        self._extract_float(item,'discount_percent'),
                                        self._extract_float(item,'original_price'), 
                                        self._extract_float(item,'price'),
                                        self._extract_float(item,'price_incl_tax'),
                                        self._extract_float(item,'qty_ordered'),
                                        self._extract_float(item,'tax_amount'), 
                                        self._extract_float(item,'row_total'), 
                                        self._extract_float(item,'row_total_incl_tax'),
                                        self._extract_field(item,'item_id'),
                                        'product',
                                    ]))    
            except:
                print("Unexpected error: " + sys.exc_info()[0])
                
        df_trans = pd.DataFrame(order_rows, columns=[
            'Transaction_ID', 
            'Increment_ID',
            'Base_Currency','Total_Gross','Discount','Sub_Total','Tax_Amount','Amount_Paid',
            'Total_Num_Items_Purchased', 'Customer_Is_Guest',
            'Email', 
            'Billing_Entity_ID', 
            'Firstname', 'Lastname', 
            'Billing_Address', 'Billing_Zip','Billing_City','Billing_Country_ID',
            'Address_Line_1','Zip','City','Country',
            'Datetime', 
            'Status', 
            'Branch_Name', 
            'Branch_ID',
            'Item_Types',
            'Voucher_Used',
            'Voucher_Amount',
            'Single_Item_Product_ID',
        ])

        df_items = pd.DataFrame(item_rows, columns=[ 
            'Transaction_ID', 'Datetime', 
            'Name', 'SKU', 'Product_ID', 
            'Discount','Discount_Percent',
            'Original_Price', 'Price', 'Price_Incl_Tax', 
            'Quantity',
            'Tax_Amount', 'Total_Minus_Tax', 'Total',
            'Item_ID',
            'Item_Type',
        ])

        return df_trans, df_items
        
        
class MagentoCustomersExtract():
    
    def __init__(self, access_token, base_url):
        
        self.api = MagentoREST(access_token, base_url)
        self.df_customers = None
        
    def _extract_field(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else ''
    
    def _extract_array(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else ['']
    
    def extract(self, page_size = 1000, verbose=False):
        
        customer_rows = []
        
        today = datetime.datetime.today()
        
        try:
            params = {'searchCriteria[currentPage]':str(1), 
                      'searchCriteria[pageSize]':str(1),
                      'searchCriteria[filter_groups][0][filters][0][field]':'created_at',
                      'searchCriteria[filter_groups][0][filters][0][value]':today.strftime('%Y-%m-%d %H:%M:%S'),
                      'searchCriteria[filter_groups][0][filters][0][condition_type]':'lteq'
                     }
            response, customers = self.api.get('customers/search', params=params, verbose=False)
            to_id = customers['total_count']
            
            if verbose:
                print("Total number of customers: {:,}".format(customers['total_count']))
        except:
            print("Unexpected error: " + sys.exc_info()[0])
        
        cid = 1
        while cid <= to_id:
                
            try:
                
                page = int(np.floor((cid-1)/float(page_size))) + 1
                if verbose:
                    print("Requesting page {} of (customers from {} to {})".format(page, (page-1)*page_size, (page-1)*page_size + page_size))
                    
                params = {'searchCriteria[currentPage]':str(page), 
                          'searchCriteria[pageSize]':str(page_size),
                          'searchCriteria[filter_groups][0][filters][0][field]':'created_at',
                          'searchCriteria[filter_groups][0][filters][0][value]':today.strftime('%Y-%m-%d %H:%M:%S'),
                          'searchCriteria[filter_groups][0][filters][0][condition_type]':'lteq'
                         }
                cid = page_size + cid
                response, customers = self.api.get('customers/search', params=params, verbose=False)

                if (200 <= response.status_code <= 299) and 'items' in customers:
                  
                    for customer in customers['items']:
                  
                        city = ''
                        country_id = ''
                        postcode = ''
                        street = ''
                        telephone = ''
                        magento_id  = self._extract_field(customer, 'id')
                        dob = self._extract_field(customer, 'dob')
                        gender = self._extract_field(customer, 'gender')
                        email = self._extract_field(customer, 'email')
                        firstname = self._extract_field(customer, 'firstname')
                        lastname = self._extract_field(customer, 'lastname')
                        group_id = self._extract_field(customer, 'group_id')
                        if 'addresses' in customer and len(customer['addresses']) > 0:
                            for addr in customer['addresses']:
                                city = self._extract_field(addr, 'city')
                                country_id = self._extract_field(addr, 'country_id')
                                postcode = self._extract_field(addr, 'postcode')
                                street = self._extract_array(addr, 'street')[0]
                                telephone = self._extract_field(addr, 'telephone')

                                customer_rows.append(tuple([
                                    magento_id,
                                    firstname, lastname,# name.strip().upper(),
                                    gender, dob, email.strip().upper(),
                                    group_id,
                                    street, postcode, city, country_id, telephone
                                ]))
                        else:
                            customer_rows.append(tuple([
                                magento_id,
                                firstname, lastname, #name.strip().upper(),
                                gender, dob, email.strip().upper(),
                                group_id,
                                street, postcode, city, country_id, telephone
                            ]))
                    
            except:
                print("Unexpected error: " + sys.exc_info()[0])

        self.df_customers = pd.DataFrame(customer_rows, columns=[
            'Magento_ID',
            'Firstname', 'Lastname', #'Name', 
            'Gender', 'DOB', 'Email',
            'Group_ID', 
            'Address_Line_1', 'Zip', 'City', 'Country', 'Telephone'
        ])
        
        return self.df_customers
        
class MagentoProductsExtract():
    
    def __init__(self, access_token, base_url):
        
        self.api = MagentoREST(access_token, base_url)
        
    def _extract_field(self, product, field_name):
        
        return product[field_name] if field_name in product else ''
    
    def _extract_array(self, product, field_name):
        
        return product[field_name] if field_name in product else []
    
    def get_additional_fields(self, product):
        
        return []
    
    def define_additional_columns(self):
        
        return []

    def extract(self, page_size = 1000, verbose=False):
        
        product_rows = []
        
        today = datetime.datetime.today()
        
        try:
            params = {'searchCriteria[currentPage]':str(1), 
                      'searchCriteria[pageSize]':str(1),
                      'searchCriteria[filter_groups][0][filters][0][field]':'created_at',
                      'searchCriteria[filter_groups][0][filters][0][value]':today.strftime('%Y-%m-%d %H:%M:%S'),
                      'searchCriteria[filter_groups][0][filters][0][condition_type]':'lteq'
                     }
            response, customers = self.api.get('products', params=params, verbose=False)
            to_id = customers['total_count']
            
            if verbose:
                print("Total number of products: {:,}".format(customers['total_count']))
        except:
            print("Unexpected error: " + sys.exc_info()[0])
        
        cid = 1
        while cid <= to_id:
                
            try:
                
                page = int(np.floor((cid-1)/float(page_size))) + 1
                if verbose:
                    print("Requesting page {} of (products from {} to {})".format(page, (page-1)*page_size, (page-1)*page_size + page_size))
                    
                params = {'searchCriteria[currentPage]':str(page), 
                          'searchCriteria[pageSize]':str(page_size),
                          'searchCriteria[filter_groups][0][filters][0][field]':'created_at',
                          'searchCriteria[filter_groups][0][filters][0][value]':today.strftime('%Y-%m-%d %H:%M:%S'),
                          'searchCriteria[filter_groups][0][filters][0][condition_type]':'lteq'
                         }
                cid = page_size + cid
                response, products = self.api.get('products', params=params, verbose=False)

                if (200 <= response.status_code <= 299) and 'items' in products:
                  
                    for product in products['items']:
        
                        #if 'product_links' in product and len(product['product_links']) > 0:
                        #    print(product['id'])
                        #    print(product['product_links'])
                        
                        prod_id = self._extract_field(product, 'id')
                        prod_type_id = self._extract_field(product, 'type_id')
                        name = self._extract_field(product, 'name')
                        sku = self._extract_field(product, 'sku')
                        price = self._extract_field(product, 'price')
                        created_at = self._extract_field(product, 'created_at')
                        updated_at = self._extract_field(product, 'updated_at')

                        #options = ','.join(self._extract_array(product, 'options'))
                        #product_links = ','.join(self._extract_array(product, 'product_links'))
                        #tier_prices = ','.join(self._extract_array(product, 'tier_prices'))
                        
                        fields = [
                            prod_id, prod_type_id, name, sku, 
                            price, 
                            created_at, updated_at
                        ]
                        
                        fields = fields + self.get_additional_fields(product)

                        product_rows.append(tuple(fields))
                            
            except:
                print("Unexpected error: " + sys.exc_info()[0])
                
        columns = [
            'Product_ID', 'Product_Type_ID', 'Name', 'SKU', 
            'Price', 
            'Created_At', 'Updated_At'
        ]
        
        columns = columns + self.define_additional_columns()
            
        df_products = pd.DataFrame(product_rows, columns=columns)
        
        return df_products
    



import binascii

class XORHasher:
    
    def __init__(self, key='12345678901234567892123456789312'):
        self.key = bytes(key, 'utf-8')
        self.n = len(self.key)
        
    def encryptInt(self, value):
        res = self.__encrdecr( bytearray(str(value).encode("utf-8")) )
        return binascii.hexlify( res ).decode('utf-8')
    
    def encryptStr(self, string):
        res = self.__encrdecr( bytearray(string, 'utf-8') )
        return binascii.hexlify( res ).decode('utf-8')
    
    def decrypt(self, hexstring):
        if len(hexstring) % 2 == 1:
            return ''
        seq = binascii.unhexlify(bytearray(hexstring, 'utf-8'))
        return self.__encrdecr( bytearray(seq) ).decode('utf-8')
    
    def __encrdecr(self, sequence):
        i = 0
        for c in sequence:
            sequence[i] = sequence[i] ^ self.key[i % self.n]
            i = 1 + i
            
        return sequence

def decrypt_columns(df, columns=[]):
    
    hh = XORHasher('SB-Humai-Fernet-Key-ABCDEFGHIJKL')
    
    for c in columns:
        df.loc[:,c] = df[c].astype(str, skipna=True)
        df.loc[:,c].fillna('', inplace=True)
        df.loc[:,c] = df[c].apply(lambda x: hh.decrypt(x)) 

def encrypt_columns(df, columns=[]):
    
    hh = XORHasher('SB-Humai-Fernet-Key-ABCDEFGHIJKL')
    
    for c in columns:
        df.loc[:,c] = df[c].astype(str, skipna=True)
        df.loc[:,c].fillna('', inplace=True)
        df.loc[:,c] = df[c].astype(str).apply(lambda x: hh.encryptStr(x))   



import pandas as pd
import numpy as np
from scipy import stats
import datetime
import os
import math
import matplotlib.pyplot as plt 

from libraries.cloud_tools import loadDatasetToPandas
from libraries.hasher import XORHasher
from libraries.product_category_model import HumaiProdCategoryFeatureGen, HumaiProdCategoryModel
from libraries.proportional_hazards_model import HumaiPropHazardsFeatureGen, HumaiPropHazardsModel

import sklearn

class HumaiProductLaunchCustomerRanking:

    def __init__(self):
        return

    def getRanking(self, new_item_name, launch_date, 
                   model,       # item name to category model
                   model_feature_gen,
                   df_prod,     # item name and category table
                   df_cds_all,  # customer discount sensitivity table
                   df_cb_all,   # customer behaviour table
                   df_items,    # table of all items in all transactions
                   phm,         # prop haz model
                   ph_feature_gen,  # feature generator used for Prop Haz model
                   use_discount_sensitivity = False,
                   use_score_to_launch = True,
                   max_days_before_launch = 30,
                   max_days_after_launch = 0,
                   score_to_launch_max_days = 30,
                   cid_col = 'Customer_ID',
                   item_name_col = 'item_name',
                   category_col = 'category',
                   verbose = True):

        # Convert item names to upper-case
        #df_prod = df_prod[[item_name_col, category_col]]
        #df_prod.dropna(inplace=True)
        #df_prod.drop_duplicates(subset=[item_name_col], inplace=True)
        #df_prod[item_name_col] = df_prod[item_name_col].apply(lambda x: x.upper())
        
        # 1-gram product name vectoriser
        #feature_gen = HumaiProdCategoryFeatureGen()
        #feature_gen.init(df_prod, item_name=item_name_col, category=category_col)

        df_items[item_name_col] = df_items[item_name_col].apply(lambda x: x.upper())

        # vectorise the new item name
        new_item_vec = model_feature_gen.generate( [new_item_name] )[0]
        print(new_item_vec)

        # Get category prediction for new item
        category = model.predict(new_item_vec)
        if verbose:
            print("Category Prediction: {0}".format(category[0]))

        dfg_prod = df_prod.groupby(category_col)

        # Keep only those items that are in the predicted category, and rank them by
        # how similar they are to the new one in terms of vector distance.

        def similarityKey(elem):
            return elem[1]

        item_similarity_score = []
            
        for category_name, df_category in dfg_prod:
            if category == category_name:
                item_names = df_category[item_name_col].tolist()
                items_vec_matrix = model_feature_gen.generate( item_names )
                
                similarities = sklearn.metrics.pairwise.cosine_similarity(items_vec_matrix, 
                                                                        new_item_vec)
                
                item_similarity_score = [ (name,sim[0]) for (name,sim) in zip(item_names, similarities) ]
                item_similarity_score.sort(key=similarityKey)
                #print(item_names)
                
                item_similarity_score = dict(item_similarity_score)

        # Discard all behaviour that doesn't involve the predicted category
        dfg_cb_all = df_cb_all.groupby(category_col)
        df_cb = pd.DataFrame()

        for category_name, df_category in dfg_cb_all:
            if category == category_name:
                
                df_cb = df_category[[cid_col, 'Num_Items_Purchased', item_name_col, 'Datetime']]

        #df_cb.Datetime = pd.to_datetime(df_cb.Datetime, format='%Y-%m-%dT%H:%M:%S.000Z')

        df_cb.Item_Name = df_cb.Item_Name.apply(lambda name: name.upper())

        #df_items['Datetime'] = pd.to_datetime(df_items['Datetime'], format='%Y-%m-%dT%H:%M:%S.000Z')
        df_items[item_name_col] = df_items[item_name_col].apply(lambda x: x.upper())

        # Calculate first launch date for each item
        dfg_items = df_items.groupby(item_name_col)

        launch_dates = []
        for item_name, df_dates in dfg_items:
            #print('Item Name: {0}, first purchase date: {1}'.format(item_name, df_dates['Datetime'].min()))
            launch_dates.append( (item_name, df_dates['Datetime'].min()) )
            
        launch_dates = dict(launch_dates)

        # Create score_to_launch for customer product purchases: 
        # The closer to initial launch date for a given product, the higher the number (0,1]
        # 1/(days_between_purchase and product launch)

        def get_score_to_launch(row):
            return 1.0 / (1.0 + (row['Datetime'] - launch_dates[row[item_name_col]]).total_seconds() / (3600*24))

        if use_score_to_launch:
            df_cb['Score_To_Launch'] = df_cb.apply(get_score_to_launch, axis=1)
        else:
            df_cb['Score_To_Launch'] = 1.0

        if verbose:
            print(df_cb['Score_To_Launch'].describe())

        # Create item_similarity_score for customer product pucrhases.
        def get_item_similarity(row):
            return item_similarity_score.get(row[item_name_col], 0.0)

        df_cb['Item_Similarity_Score'] = df_cb.apply(get_item_similarity, axis=1)

        if verbose:
            print(df_cb['Item_Similarity_Score'].describe())

        if use_score_to_launch:
            score_to_launch_max_days = 30

            # Filter customers and items, based on score_to_launch cutoff.

            # This gets rid of customers that have never bought near the launch date,
            # as well as items that specific customers have never bought near a launch date.
            df_cb.drop( df_cb[df_cb['Score_To_Launch'] < 1.0/score_to_launch_max_days ].index, inplace=True )
            len(df_cb)

        def get_discount_sens(row):
            return df_cds_all[ df_cds_all[cid_col] == row[cid_col] ].iloc[0,:].Discount_Sensitivity

        if use_discount_sensitivity:
            df_cb['Discount_Sensitivity'] = df_cb.apply(get_discount_sens, axis=1)
        else:
            df_cb['Discount_Sensitivity'] = 100.0

        # For each row, calculate score:
        # (Total_Num_Items_Purchased for customer) * item_similarity_score * score_to_launch

        def calc_score(row):
            return row['Item_Similarity_Score'] * row['Score_To_Launch'] * row['Discount_Sensitivity']

        df_cb['Final_Score'] = df_cb.apply(calc_score, axis=1)

        if verbose:
            print(df_cb['Final_Score'].describe())

        # Get ranking, then get unique customers
        df_cb.sort_values(by=['Final_Score'], ascending=False, inplace=True)
        df_cb.drop_duplicates(subset=[cid_col], inplace=True)

        df_to_pred = ph_feature_gen.dataset.loc[ph_feature_gen.dataset[cid_col].isin(df_cb[cid_col])]

        # Predictions
        pred, dates = phm.predict(df_to_pred, 
                        features=None,
                        threshold=0.7)

        send_dates = pd.DataFrame()
        send_dates[cid_col] = df_to_pred[cid_col]
        send_dates['Datetime'] = dates

        if verbose:
            print(send_dates['Datetime'].head())

        date_col = 'Send_Mail_Datetime'
        df_cb[date_col] = launch_date
        # Set predicted send date for all customers for which we could predict it.
        # For all othe customers, set predicted send date to be the launch date.
        mask = df_cb[df_cb[cid_col].isin(send_dates[cid_col])].index
        df_cb.loc[mask,date_col] = pd.to_datetime(send_dates['Datetime'].values)

        if verbose:
            print(df_cb[date_col].head())

        earliest_mail_date = launch_date - datetime.timedelta(days=max_days_before_launch)
        latest_mail_date = launch_date + datetime.timedelta(days=max_days_after_launch)
        df_cb[date_col] = df_cb[date_col].apply(lambda x: min(latest_mail_date, earliest_mail_date if x <= earliest_mail_date else x) )

        df_cb = df_cb[['Email', 'Final_Score', 'Send_Mail_Datetime']]

        return df_cb





import pandas as pd
import numpy as np
import datetime

from pathlib import Path

from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, TransformerMixin
import joblib

from humailib.cloud_tools import GoogleCloudStorage
from humailib.transformers import DataframeFeatureUnion

class PipelineBuilder:
    """
    This class allows you to build a pipeline, by executing and adding one ore more steps 
    at a time, in a way that lends itself well to how data transformation is implemented in Notebooks.
    """
    
    def __init__(self):
        
        self.steps = {}
        self.step_names = {}
        
        self.new_pipeline('default')
        
        return
    
    def new_pipeline(self, name):
        
        self.steps[name] = []
        self.step_names[name] = {}
        self.current_pipeline = name
    
    def transform_and_add(self, steps, df):
        
        pipeline = Pipeline(steps)
        df_out = pipeline.transform(df)
        
        self.add(steps)
        
        return df_out
    
    def fit_transform_and_add(self, steps, df):
        
        pipeline = Pipeline(steps)
        df_out = pipeline.fit_transform(df)
        
        self.add(steps)
        
        return df_out
    
    def add(self, steps):
        
        for step in steps:
            if step[0] not in self.step_names[self.current_pipeline]:
                self.steps[self.current_pipeline].append(step)
                self.step_names[self.current_pipeline][step[0]] = len(self.steps[self.current_pipeline])-1
                
        return
    
    def add_pipeline(self, pipeline):
        
        for name, sub in pipeline.steps:
            self.steps[ name ] = sub.steps
            self.step_names[ name ] = {sub_step_name:i for i,(sub_step_name,_) in enumerate(sub.steps)}
            
        return
    
    
    def get_all_pipelines(self):
        
        return [pipeline_name for pipeline_name in self.steps.keys() if len(self.steps[pipeline_name]) > 0]
            
    
    def build(self, pipelines=None, catch_all_params=None, overwrite_params=False, verbose=False):
        
        assert pipelines is not None and len(pipelines)>0
        
        if isinstance(pipelines, str):
            pipelines = [pipelines]
        
        sub_names = []
        sub_pipes = []
        for current_pipeline in pipelines:
            if current_pipeline in self.steps and len(self.steps[current_pipeline]) > 0:
                sub_names.append(current_pipeline)
                sub_pipes.append(self.steps[current_pipeline])
                
        pipe_out = Pipeline([
            (name, Pipeline(sub)) for name,sub in zip(sub_names,sub_pipes)
        ])
        
        if verbose:
            print("[PipelineBuilder::build] Built:")
            print(pipe_out.steps)
            
        params_changed = False        
        if catch_all_params is not None:
            for catch_all_name, catch_all_value in catch_all_params.items():
                for current_pipeline in pipelines:
                    # Get the parameters for all components in this pipeline.
                    components_params = pipe_out.named_steps[current_pipeline].get_params()
                    if verbose:
                        print("[PipelineBuilder::build] ==== All component params for '{}':".format(current_pipeline))
                        for p,v in components_params.items():
                            print('  {} -> {}'.format(p,v))
                            
                    # See if any parameters have names that contain the catch-all name
                    params = {name:catch_all_value for name,_ in components_params.items() if catch_all_name in name}
                    if not overwrite_params:
                        for name,val in components_params.items():
                            if catch_all_name in name:
                                if isinstance(val, list):
                                    params[name] = list(set(val+catch_all_value))
                                elif isinstance(val, dict):
                                    for k,v in catch_all_value.items():
                                        val[k] = v
                                    params[name] = val
                    if verbose:
                        print("[PipelineBuilder::build] ==== Found matching component params:")
                        for p,v in params.items():
                            print('  {} -> {}'.format(p,v))
                            
                    # Set these to have the catch-all value
                    if len(params) > 0:
                        print("[PipelineBuilder::build] ==== Setting params...".format(params))
                        for p,v in params.items():
                            if isinstance(v, str):
                                print("...setting '{}' -> '{}'".format(p,v))
                            else:
                                print("...setting '{}' -> {}".format(p,v))
                        params_changed = True
                        pipe_out.named_steps[current_pipeline].set_params(**params)
                    
        if not params_changed:
            print("[PipelineBuilder::build] No params changed.")
        
        return pipe_out
    
    def load(self, filename, gcs=None, cache_dir='./cache'):
        
        if 'gs://' in filename:
            cache_file = Path(cache_dir + '/' + filename.split('/')[-1])
            gcs.download_file(filename, cache_file)
            pipeline = joblib.load(cache_file)
        else:
            pipeline = joblib.load(filename)
            
        self.__init__()
        self.add_pipeline(pipeline)
            
        return
    
    def save(self, filename, gcs=None, cache_dir='./cache'):
        
        if len(self.steps) == 0:
            return
        
        pipeline = self.build(pipelines=[name for name in self.step_names])
        
        if 'gs://' in filename:
            if gcs is None:
                gcs = GoogleCloudStorage()
            cache_file = cache_dir + '/' + filename.split('/')[-1]
            #print(cache_file)
            joblib.dump(pipeline, cache_file)
            gcs.upload_file(cache_file, filename)
        else:
            joblib.dump(pipeline, filename)
        
        return
    


from setuptools import setup
from setuptools import find_packages

REQUIRED_PACKAGES = [
    'pandas-gbq>=0.13.2',
    'pathlib',
    'google-cloud-datastore','google-cloud-bigquery','google-cloud-storage','tqdm','google-cloud-bigquery-storage',
    'pyarrow',
    'pandas==1.0.0'
]

setup(
    name='humailib',
    version='1.0.32',
    description='HUMAI framework package',
    url='',
    author='HUMAI',
    author_email='research@humai.nl',
    license='',
    install_requires=REQUIRED_PACKAGES,
    packages=find_packages(),
    zip_safe=False
)


# -*- coding: utf-8 -*-
""" transformers.py

A collection of often-used sklearn custom Transformers.

"""

import pandas as pd
import numpy as np
import inspect, sys
import datetime

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import FeatureUnion, _fit_transform_one, _transform_one
from joblib import Parallel, delayed

from scipy import sparse

import humailib.utils as hu


def get_all_transformers():
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            print(obj)


class DataframeFunctionTransformer(BaseEstimator, TransformerMixin):
    """
    Transform dataframe with a custom function.
    The function, func, needs to be of the format:
      func(X, params)
    """
    def __init__(self, func, params):
        self.func = func
        self.params = params

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None, copy=None):
        
        #assert isinstance(input_df, pd.DataFrame) or isinstance(input_df, pd.Series)
        
        print("[DataframeFunctionTransformer] with function '{}' on {:,} rows...".format(self.func.__name__, len(X)))
        print("  params={}".format(self.params))
        return self.func(X, self.params)
    
    
class DataframeFeatureUnion(FeatureUnion):
    """
    FeatureUnion that works and preserves pandas Dataframes.
    From: https://zablo.net/blog/post/pandas-dataframe-in-scikit-learn-feature-union/
    
    If using PipelineBuilder, treat a DataframeFeatureUnion as a transformer step in the
    pipeline.
    
    Example:
    
        pipeline_builder = PipelineBuilder()
        pipeline_builder.new_pipeline('aggregate_and_finalise')        
        ...
        
        df_aggregate = pipeline_builder.transform_and_add([
            ('aggregate_customers', ht.DataframeFeatureUnion([
                ('agg_max', ht.Aggregator(
                    agg_type='max', agg_column='Customer_ID', columns=['A','B'],
                    copy_agg_column=True
                )), 
                ('agg_mean', ht.Aggregator(
                    agg_type='mean', agg_column='Customer_ID', columns=['C','D'],
                    copy_agg_column=False
                )),
                ('agg_sum', ht.Aggregator(
                    agg_type='sum', agg_column='Customer_ID', columns=['E','F'],
                    copy_agg_column=False
                )) 
            ], n_jobs=3)
            )
        ], df)

    """
    
    def fit_transform(self, X, y=None, **fit_params):
        self._validate_transformers()
        result = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_transform_one)(trans, weight=weight, X=X, y=y,
                                        **fit_params)
            for name, trans, weight in self._iter()
        )

        if not result:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        
        Xs, transformers = zip(*result)
        self._update_transformer_list(transformers)
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = self.merge_dataframes_by_column(Xs)
            
        return Xs

    def merge_dataframes_by_column(self, Xs):
        return pd.concat(Xs, axis="columns", copy=False)

    def transform(self, X):
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(trans, weight=weight, X=X, y=None)
            for name, trans, weight in self._iter()
        )
        
        if not Xs:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = self.merge_dataframes_by_column(Xs)
            
        return Xs
    

class CrossReferenceTransformer(BaseEstimator, TransformerMixin):
    """
    Cross reference a column in X with the same column in a reference
    dataframe, and return the values from the return column.
    
    Parameters
    ----------
    df_ref : The reference data frame that contains the reference column
             as well as the return column
    ref_column : The reference column that is present in both the dataframe
                 as well as the reference frame
    ret_column : Which column to return from the reference frame
    na_value : Any missing values will be replaced with this value
    """
    def __init__(self, df_ref, ref_column, ret_column, na_value=np.nan):
        self.df_ref = df_ref
        self.ref_column = ref_column
        self.ret_column = ret_column
        self.na_value = na_value
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[CrossReferenceTransformer] transform on {:,}...".format(len(X)))
        print("  Cross-referencing {}, and returning {}".format(self.ref_column, self.ret_column))
        
        a_to_b = {a:b for a,b in zip(self.df_ref[self.ref_column].to_numpy(), self.df_ref[self.ret_column].to_numpy())}
    
        df = X.copy()
        df.loc[:,self.ret_column] = X.apply(
            lambda x: a_to_b.get(x[self.ref_column], self.na_value), 
            axis=1
        )
        
        return df
    

class KeepMostRecentTransformer(BaseEstimator, TransformerMixin):
    """
    Keep the most recent n rows for each entity of [agg_column]
    """
    def __init__(self, agg_column, datetime_column, columns=None, n=5):
        self.agg_column = agg_column
        self.datetime_column = datetime_column
        self.columns = columns
        self.n = n
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[KeepMostRecentTransformer] transform on {:,}...".format(len(X)))
        print("  agg_column={}, columns={}, n={}".format(self.agg_column, self.columns, self.n))
        
        df = X.copy()

        df.loc[:,'kmr_ranking'] = df.sort_values(
            by=[self.agg_column, self.datetime_column]
        ).groupby(by=self.agg_column)[self.datetime_column].rank(method='first', ascending=False)

        df = df[df['kmr_ranking'] <= self.n]
        
        print("[KeepMostRecentTransformer] left {:,} rows, or {:,} {}s.".format(len(df), df[self.agg_column].nunique(), self.agg_column))

        if self.columns:
            df = df[self.columns]
        else:
            del df['kmr_ranking']

        return df
    
    
class ColumnStatsTransformer(BaseEstimator, TransformerMixin):
    """
    Print out exploratory information for certain columns in a dataframe. 
    """
    def __init__(self, columns=None, max_rows=5):
        self.columns = columns
        self.max_rows = max_rows
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[ColumnStatsTransformer] stats {}...".format(self.columns))
        
        if self.columns is not None:
            for c in self.columns:
                hu.column_stats(X, c, n=self.max_rows)
            
        return X 
           
        
class SelectColumnsTransformer(BaseEstimator, TransformerMixin):
    """
    Select which columns to keep, and discard the rest.
    """
    def __init__(self, columns=None):
        self.columns = columns

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[SelectColumnsTransformer] select {}...".format(self.columns))
        cpy_df = X[self.columns].copy()
        return cpy_df


class DropColumnsTransformer(BaseEstimator, TransformerMixin):
    """
    Drop certain columns from a dataframe.
    """
    def __init__(self, columns=None):
        self.columns = columns

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DropColumnsTransformer] drop {}...".format(self.columns))
        
        return X.drop(columns=self.columns)
    
    
class DropDuplicatesTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, subset=None):
        self.subset = subset

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DropDuplicatesTransformer] dropping duplicates of {} on {:,}...".format(self.subset, len(X)))
    
        df = X.drop_duplicates(subset=self.subset)
    
        print("[DropDuplicatesTransformer] left: {:,}".format(len(df)))
    
        return df
    
    
class DropNATransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, subset=None):
        self.subset = subset

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DropNATransformer] dropping NaNs of {} on {:,}...".format(self.subset, len(X)))
    
        df = X.dropna(subset=self.subset)
    
        print("[DropNATransformer] left: {:,}".format(len(df)))
    
        return df  
    
    
class DropEmptyStringTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, column=None):
        self.column = column

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DropEmptyStringTransformer] dropping empty strings of {} on {:,}...".format(self.column, len(X)))
    
        df = X.copy()
        
        df.drop(index=df[df[self.column].str.strip() ==''].index, inplace=True)
    
        print("[DropEmptyStringTransformer] left: {:,}".format(len(df)))
    
        return df  
    
    
class SortingTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, columns=None):
        self.columns = columns

    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[SortingTransformer] sorting on {} on {:,}...".format(self.columns, len(X)))
    
        df = X.sort_values(by=self.columns).copy()
    
        return df
    
    
class IntTypeTransformer(BaseEstimator, TransformerMixin):
    
    """
    """ 
    def __init__(self, columns = None, na_replace_value=None, drop_na=False):
        self.columns = columns
        self.na_replace_value = na_replace_value
        self.drop_na = drop_na
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[IntTypeTransformer] transforming {} on {:,}...".format(self.columns, len(X)))
        print("  na_replace_value={}, drop_na={}".format(self.na_replace_value, self.drop_na))

        df = X.copy()
        hu.columns_as_int(df, columns=self.columns, na_replace_value=self.na_replace_value, drop_na=self.drop_na)

        print("[IntTypeTransformer] left: {:,}".format(len(df)))

        return df

    
class StrTypeTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, columns = None, uppercase=False, drop_empty=False):
        self.columns = columns
        self.uppercase = uppercase
        self.drop_empty = drop_empty
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[StrTypeTransformer] transform {} on {:,}...".format(self.columns, len(X)))
        print("  uppercase={}, drop_empty={}".format(self.uppercase, self.drop_empty))

        df = X.copy()
        hu.columns_as_str(
            df, columns=self.columns, uppercase=self.uppercase, drop_empty=self.drop_empty
        )
        
        if self.drop_empty:
            print("[StrTypeTransformer] after dropping '': {:,}".format(len(df)))
        
        return df

    
class DatetimeTypeTransformer(BaseEstimator, TransformerMixin):
    """
    """
    def __init__(self, columns = None, datetime_format='%Y-%m-%d %H:%M:%S'):
        self.columns = columns
        self.datetimeformat = datetimeformat
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[DatetimeTypeTransformer] transform {} on {:,}...".format(self.columns, len(X)))
        print("  datetimeformat={}".format(self.datetimeformat))

        df = X.copy()
        hu.columns_as_datetime(df, columns=self.columns, datetime_format=self.datetimeformat)
        
        return df
    
"""
"""    
class TransactionsCleanupTransformer(BaseEstimator, TransformerMixin):
    
    def __init__(self, datetime_column = 'Datetime', tid_column = 'Transaction_ID', cid_column = 'Customer_ID', datetime_format='%Y-%m-%d %H:%M:%S'):
        self.datetime_column = datetime_column
        self.tid_column = tid_column
        self.cid_column = cid_column
        self.datetime_format = datetime_format
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[TransactionsCleanupTransformer] transform on {:,}...".format(len(X)))
        
        print("  datetime_column={}, tid_column={}, cid_column={}".format(
            self.datetime_column, self.tid_column, self.cid_column
        ))

        df = X.copy()
        hu.transactions_cleanup(
            df, 
            datetime_column = self.datetime_column, 
            tid_column = self.tid_column, 
            cid_column = self.cid_column, 
            datetime_format=self.datetime_format
        )
        
        return df
    
    
class KeepEmailActivitiesTransformer(BaseEstimator, TransformerMixin):
    """
    """
    
    def __init__(self, activities):
        self.activities = activities
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[KeepEmailActivitiesTransformer] keeping {} on {:,}...".format(self.activities, len(X)))

        df = X.copy()
        hu.email_keep_activities(
            df,
            activities=self.activities
        )
        
        print("[KeepEmailActivitiesTransformer] left {:,}".format(len(df)))
        
        return df
    

class UniqueEmailEventsTransformer(BaseEstimator, TransformerMixin):
    """
    """
    
    def __init__(self, cid_column = 'Customer_ID', activity_column = 'Activity', activity_datetime_column = 'Activity_Datetime', verbose=True):
        self.cid_column = cid_column
        self.activity_column = activity_column
        self.activity_datetime_column = activity_datetime_column
        self.verbose = verbose
        return
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[UniqueEmailEventsTransformer] transform on {:,}...".format(len(X)))
        print("  cid_column={}, activity_column={}, activity_datetime_column={}".format(
            self.cid_column, self.activity_column, self.activity_datetime_column
        ))

        df = X.copy()
        hu.email_ensure_zero_or_one_activity_per_broadcast(
            df, 
            cid_column = self.cid_column,
            activity_column = self.activity_column,
            activity_datetime_column = self.activity_datetime_column,
            verbose=self.verbose
        )
        
        print("[UniqueEmailEventsTransformer] left {:,}".format(len(df)))
        
        return df
    
    
    
class SeasonalityTransformer(BaseEstimator, TransformerMixin):
    
    """
    Calculate seasonality labels (one-hot encoding type) for each row
    
    Input dataframes could be:
        1. Items (to get product seasonality)
        2. Transactions (to get customer seasonality)
    """
    
    def __init__(self, datetime_column = 'Datetime'):
        self.datetime_column = datetime_column
        #print(self.datetime_column)
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[SeasonalityTransformer] transform on {:,}...".format(len(X)))
        
        df = X.copy()
        #print(df.head())
    
        print("  spring...")
        df.loc[:,'spring'] = df.apply(
            lambda x: 1.0 if (4 <= x[self.datetime_column].month < 7) else 0.0, axis=1
        )

        print("  summer...")
        df.loc[:,'summer'] = df.apply(
            lambda x: 1.0 if (7 <= x[self.datetime_column].month < 10) else 0.0, axis=1
        )

        print("  autumn...")
        df.loc[:,'autumn'] = df.apply(
            lambda x: 1.0 if (10 <= x[self.datetime_column].month <= 12) else 0.0, axis=1
        )

        print("  winter...")
        df.loc[:,'winter'] = df.apply(
            lambda x: 1.0 if (1 <= x[self.datetime_column].month < 4) else 0.0, axis=1
        )
        
        print("[SeasonalityTransformer] Done.")

        return df
    

class TimeOfDayTransformer(BaseEstimator, TransformerMixin):
    
    """
    Calculate email activity time-of-day (one-hot encoding type) for each row
    
    Input dataframe:
      1. Email events
    """
    
    def __init__(self, datetime_column = 'Datetime'):
        self.datetime_column = datetime_column
        #print(self.datetime_column)
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[TimeOfDayTransformer] transform on {:,}...".format(len(X)))
        
        df = X.copy()
        #print(df.head())
    
        print("  morning...")
        df.loc[:,'morning'] = df.apply(
            lambda x: 1.0 if (0 <= x[self.datetime_column].hour < 12) else 0.0, axis=1
        )

        print("  daytime...")
        df.loc[:,'daytime'] = df.apply(
            lambda x: 1.0 if (12 <= x[self.datetime_column].hour < 17) else 0.0, axis=1
        )

        print("  evening...")
        df.loc[:,'evening'] = df.apply(
            lambda x: 1.0 if (17 <= x[self.datetime_column].hour <= 23) else 0.0, axis=1
        )
        
        print("[TimeOfDayTransformer] Done.")

        return df



class GenderTransformer(BaseEstimator, TransformerMixin):
    
    """
    Calculate gender labels (one-hot encoding type) for each row
    
    Input dataframes could be:
        1. Items (to get product 'gender', i.e. what proportion of the product is bought by male, female, unisex)
    """
    
    def __init__(self, df_trans, df_customers, cid_column = 'Customer_ID', gender_column = 'Gender'):
        self.cid_to_gender = {cid:gender for cid,gender in zip(df_customers[cid_column].values, df_customers[gender_column].values) }
        self.tid_to_gender = {tid:self.cid_to_gender[cid] for tid,cid in zip(df_trans['Transaction_ID'].values, df_trans[cid_column].values)}
        self.cid_column = cid_column
        #print(self.datetime_column)
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[ProductGenderTransformer] transform on {:,}...".format(len(X)))
        
        df = X.copy()
        
        df.loc[:,'male'] = df['Transaction_ID'].apply(
            lambda x: 1.0 if (x in self.tid_to_gender and self.tid_to_gender[x] == 'M') else 0.0, 
                                   axis=1
        )
        df.loc[:,'female'] = df['Transaction_ID'].apply(
            lambda x: 1.0 if (x in self.tid_to_gender and self.tid_to_gender[x] == 'F') else 0.0, 
                                   axis=1
        )
        df.loc[:,'unisex'] = df['Transaction_ID'].apply(
            lambda x: 1.0 if (x in self.tid_to_gender and (self.tid_to_gender[x] != 'M' and self.tid_to_gender[x] != 'F')) else 0.0, 
                                   axis=1
        )        

        return df

    
class Aggregator(BaseEstimator, TransformerMixin):
    """
    """ 
    def __init__(self, agg_type, agg_column, columns, **kwargs):
        self.agg_type = agg_type
        self.agg_column = agg_column
        self.columns = columns
        self.kwargs = kwargs
        return
    
    def fit(self, X, y=None):
        return
    
    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        assert self.agg_column in X, "{} does not exist!!".format(self.agg_column)
        
        n = X[self.agg_column].nunique()
        print("[Aggregator] transform on {:,}...".format(n))
        print("  method='{}', agg_column='{}', columns={}".format(
            self.agg_type, self.agg_column, self.columns
        ))
        for c in self.columns:
            if c not in self.columns:
                print("[Aggregator] '{}' does not exist!!".format(c))
        
        df = hu.aggregate_columns(
            X.copy(),
            self.agg_type, 
            self.agg_column,
            self.columns,
            **self.kwargs
        )
        
        print("[Aggregator] Done.")
        
        return df
    
    
class AggregatorMean(Aggregator):
    """
    """ 
    def __init__(self, agg_column, columns, copy_agg_column=True):
        super.__init__('mean', agg_column, columns, copy_agg_column)
        return
    
        
class OrderSegmentTransformer(BaseEstimator, TransformerMixin):
    """
    """        
    
    def __init__(self, num_trans_column = 'n_trans', num_segments=3):
        self.num_trans_column = num_trans_column
        self.num_segments = num_segments
        return
    
    def fit(self, X, y=None):
        return
    
    def _get_segment(self, x):
        if x.n_trans >= self.num_segments:
            return 'os_{}+'.format(self.num_segments)
        
        return 'os_{}'.format(x.n_trans)

    def transform(self, X):
        
        assert isinstance(X, pd.DataFrame)
        
        print("[OrderSegmentTransformer] transform on {:,}...".format(len(X)))

        df = X.copy()
        
        df.loc[:,'order_segment'] = df.apply(lambda x: self._get_segment(x), axis=1)
        
        return df
    
    



import pandas as pd
import numpy as np

def GetWeekdaySequencesIncNonOpens(dfg_customers, weekday_column='transaction_weekday', activity_column='activity', min_seq_len = 5, max_seq_len=15, verbose=False):
    
    num_days = 8  # 7 + non-open
    
    customers = []
    sequences = []
    seqlens = []
    
    ii = 0
    nc = len(dfg_customers)
    
    for cid, g in dfg_customers:
        
        n = 0
        label = np.zeros(num_days)
        
        if (len(g[weekday_column]) >= min_seq_len):
            
            if verbose:
                print(g[weekday_column].iloc[:].values)
                print(g[activity_column].iloc[:].values)
                
            if g[weekday_column].iloc[-max_seq_len:].isnull().any():
                print("NaN detected for customer: {0}".format(cid))
                print(g[weekday_column].iloc[:].values)
                print(g[activity_column].iloc[:].values)
                g[weekday_column].fillna(7, inplace=True)
            
            h = [7 if (a=='BROADCASTS' or x==7) else int(x) for a,x in 
                 zip(g[activity_column].iloc[-max_seq_len:].values, 
                     g[weekday_column].iloc[-max_seq_len:].values)]
            
            # If the sequence consists of nothing but non-opens, make
            # its length 0.
            if np.min(h) == 7:
                n = 1 #0
            else:
                n = len(h)
                
            # If h is less than [time_steps] in length, pad it with last element
            if len(h) < max_seq_len:
                
                nn = max(0, max_seq_len - len(h))
                h.extend([h[-1] for i in range(0,nn)])
                
            #if n > 0:
            if verbose:
                print("h={0} (n={1})".format(h, n))
                
            customers.append(cid)
            sequences.append(np.array(h))
            seqlens.append(n)
        
        ii = 1 + ii
        
        if ii % 1000 == 0:
            print("  {0}/{1}".format(ii, nc))   
        
    return customers, sequences, seqlens


def GetWeekdaySequencesAndLabelsIncNonOpens(dfg_customers, weekday_column='transaction_weekday', activity_column='activity', min_seq_len = 5, max_seq_len=15, verbose=False):
    
    num_days = 8  # 7 + non-open
    
    customers = []
    sequences = []
    labels = []
    labelid = []
    seqlens = []
    
    ii = 0
    nc = len(dfg_customers)

    for cid, g in dfg_customers:

        
        n = 0
        label = np.zeros(num_days)
        
        if (len(g[weekday_column]) > min_seq_len):
            
            if verbose:
                print(g[weekday_column].iloc[:].values)
                print(g[activity_column].iloc[:].values)
                
            if g[weekday_column].iloc[-(max_seq_len+1):-1].isnull().any():
                print("NaN detected for customer: {0}".format(cid))
                print(g[weekday_column].iloc[:].values)
                print(g[activity_column].iloc[:].values)
                g[weekday_column].fillna(7, inplace=True)
            
            h = [7 if (a=='BROADCASTS' or x==7) else int(x) for a,x in 
                 zip(g[activity_column].iloc[-(max_seq_len+1):-1].values, 
                     g[weekday_column].iloc[-(max_seq_len+1):-1].values)]
                
            l = int(g[weekday_column].iloc[-1:].values[0]) if g[activity_column].iloc[-1] != 'BROADCASTS' else 7
            label[ l ] = 1
            
            # If the sequence consists of nothing but non-opens, make
            # its length 0.
            if np.min(h) == 7:
                n = 1 #0
            else:
                n = len(h)
            
            # If h is less than [time_steps] in length, pad it with last element
            if len(h) < max_seq_len:
                
                nn = max(0, max_seq_len - len(h))
                h.extend([h[-1] for i in range(0,nn)])
                    
            #if n > 0:
            if verbose:
                print("h={} -> {}: {} (n={})".format(h, l, label, n))
                
            customers.append(cid)
            sequences.append(np.array(h))
            seqlens.append(n)
            labels.append(np.array(label))
            labelid.append(l)

        ii = 1 + ii
        
        if ii % 1000 == 0:
            print("  {0}/{1}".format(ii, nc))
        
    return customers, sequences, labels, labelid, seqlens


def sqm_create_training_data(deliveries, 
                             weekday_column='transaction_weekday', 
                             activity_column='activity', 
                             min_seq_len = 5, 
                             max_seq_len=15, 
                             verbose=False):


    return



import pandas as pd
import numpy as np

import humailib.hasher as hh

class ProductCharacteristicsHelper:
    
    def __init__(self, gbq, table, product_column = 'Product_ID'):
        
        self.gbq = gbq
        self.table = table
        self.product_column = product_column
        self.df = self.__download()
        return
    
    def __download(self):
        
        self.df = self.gbq.download_table_to_pandas(self.table)
        return self.df
    
    def upload(self):
        
        self.gbq.upload_and_replace_table(self.df, self.table)
    
    def add_or_replace_columns(self, df, columns=None):
        
        if columns is None:
            columns = [c for c in df.columns if c != self.product_column]
            
        if self.df is None:
            print("[ProductCharacteristicsHelper::add_columns] table {} does not exist, creating...".format(self.table))
            self.df = df[columns].copy()
        else:
            print("[ProductCharacteristicsHelper::add_column] table {} exists, merging...".format(self.table))
            for c in columns:
                if c in self.df:
                    del self.df[c]

            self.df = self.df.merge(df[[self.product_column] + columns], on=self.product_column, how='outer')
            
        return self.df
    
    
class CustomerCharacteristicsHelper:
    
    def __init__(self, gbq, table, cid_column = 'Customer_ID', encryption = False):
        
        self.gbq = gbq
        self.table = table
        self.cid_column = cid_column
        self.encryption = encryption
        self.df = self.__download()
        return
    
    def __download(self):
        
        self.df = self.gbq.download_table_to_pandas(self.table)
        if self.df is not None and self.encryption:
            hh.decrypt_columns(self.df, columns=[self.cid_column])
        return self.df
    
    def upload(self):
        
        if self.df is not None and self.encryption:
            hh.encrypt_columns(self.df, columns=[self.cid_column])
        self.gbq.upload_and_replace_table(self.df, self.table)
    
    def add_or_replace_columns(self, df, columns=None):
        
        if columns is None:
            columns = [c for c in df.columns if c != self.cid_column]
            
        if self.df is None:
            print("[CustomerCharacteristicsHelper::add_columns] table {} does not exist, creating...".format(self.table))
            self.df = df[[self.cid_column] + columns].copy()
        else:
            print("[CustomerCharacteristicsHelper::add_column] table {} exists, merging...".format(self.table))
            for c in columns:
                if c in self.df:
                    del self.df[c]

            self.df = self.df.merge(df[[self.cid_column] + columns], on=self.cid_column, how='outer')
            
        return self.df
            
            
        



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import tensorflow as tf
from tensorflow.python.framework import dtypes
from tensorflow.contrib import learn as tflearn
from tensorflow.contrib import layers as tflayers


def lstm_model(num_units, rnn_layers, dense_layers=None, learning_rate=0.1, optimizer='Adagrad'):
    """
    Creates a deep model based on:
        * stacked lstm cells
        * an optional dense layers
    :param num_units: the size of the cells.
    :param rnn_layers: list of int or dict
                         * list of int: the steps used to instantiate the `BasicLSTMCell` cell
                         * list of dict: [{steps: int, keep_prob: int}, ...]
    :param dense_layers: list of nodes for each layer
    :return: the model definition
    """

    def lstm_cells(layers):
        if isinstance(layers[0], dict):
            return [tf.contrib.rnn.DropoutWrapper(
                tf.contrib.rnn.BasicLSTMCell(
                    layer['num_units'], state_is_tuple=True
                ),
                layer['keep_prob']
            ) if layer.get('keep_prob') else tf.contrib.rnn.BasicLSTMCell(
                    layer['num_units'],
                    state_is_tuple=True
                ) for layer in layers
            ]
        return [tf.contrib.rnn.BasicLSTMCell(steps, state_is_tuple=True) for steps in layers]

    def dnn_layers(input_layers, layers):
        if layers and isinstance(layers, dict):
            return tflayers.stack(input_layers, tflayers.fully_connected,
                                  layers['layers'],
                                  activation=layers.get('activation'),
                                  dropout=layers.get('dropout'))
        elif layers:
            return tflayers.stack(input_layers, tflayers.fully_connected, layers)
        else:
            return input_layers

    def _lstm_model(X, y):
        stacked_lstm = tf.contrib.rnn.MultiRNNCell(lstm_cells(rnn_layers), state_is_tuple=True)
        X = tf.cast(X,tf.float32)
        x_ = tf.unstack(X, axis=1, num=num_units)
        
        output, layers = tf.contrib.rnn.static_rnn(stacked_lstm, x_, dtype=dtypes.float32)
        output = dnn_layers(output[-1], dense_layers)
        prediction, loss = tflearn.models.linear_regression(output, y)
        train_op = tf.contrib.layers.optimize_loss(
            loss, tf.contrib.framework.get_global_step(), optimizer=optimizer,
            learning_rate=learning_rate)
        return prediction, loss, train_op

    return _lstm_model



import pandas as pd
import numpy as np
import plotly.graph_objects as go
import sklearn

def group_differences(df, threshold, covariate_columns):
    print("Above threshold group:")
    print(df[df.pred_ranking > threshold][covariate_columns].describe())
    print("\nBelow threshold group:")
    print(df[df.pred_ranking <= threshold][covariate_columns].describe())
    
def print_model_stats(df, pred_threshold, observed_column='observed', pred_column='pred_observed'):
    
    df_stats, _ = rate_plot(df, thresholds=[pred_threshold])

    total_positive = df[observed_column].sum()
    total_negative = len(df) - total_positive
    total_pred_positive = df[pred_column].sum()
    response_rate = total_positive / len(df)

    print("Total positive: {:,} out of {:,} (response rate: {:.1f}%)".format(total_positive, len(df), 100*response_rate))
    print("Total predicted positive: {:,} out of {:,} (response rate: {:.1f}%)".format(total_pred_positive, len(df), 100*(total_pred_positive/len(df))))
    print("Proportion of predicted positives actually positive (precision): {:.1f}%".format(100.0*df_stats['precision_1'][0]))
    print("Proportion of positives, predicted as such (recall): {:.1f}%".format(100.0*df_stats['recall_1'][0]))
    print("Proportion of predicted negatives actually negative (precision): {:.1f}%".format(100.0*df_stats['precision_0'][0]))
    print("Proportion of negatives, predicted as such (recall): {:.1f}%".format(100.0*df_stats['recall_0'][0]))
    #print("F1: {:.2f}".format(2*(precision*recall)/(precision+recall)))


def gains_chart(df_in, ranking='pred_ranking', response='observed', 
                title='Cumulative Gains', xaxis_title='% Customers Contacted',
                yaxis_title='% Positive Response', n=100):

    df = df_in.sample(frac=1).sort_values(by=[ranking], ascending=False)

    scale = df[response].sum()

    gains = [0.0]

    nn = len(df)
    split_n = int( np.ceil(nn / (n)) )

    ii = 0
    gain = 0
    while ii < nn:

        delta = min(ii+split_n, nn)
        
        gain = gain + df.iloc[ii:delta,:][response].sum()
        gains.append(gain)

        ii = split_n + ii     

    gains = [100.0*g / scale for g in gains]
    first_n = len(gains)

    #for i,j in zip(cum_gains, np.linspace(0,1,first_n)):
    #    print(i,j)

    fig = go.Figure()

    fig.add_trace(go.Scatter(x=np.linspace(0,1,first_n)*100, y=np.linspace(0,1,first_n)*100,
                            mode='lines', name='random'))
    fig.add_trace(go.Scatter(x=np.linspace(0,1,first_n)*100, y=gains,
                            mode='lines', name='model'))

    fig.update_xaxes(tickvals=np.linspace(0,100,11))
    fig.update_yaxes(tickvals=np.linspace(0,100,11))

    # Edit the layout
    fig.update_layout(title=title,
                    xaxis_title=xaxis_title,
                    yaxis_title=yaxis_title)

    df_gains = pd.DataFrame([(g,l,g/l if l > 0 else 0) for g,l in zip(gains, np.linspace(0,1,first_n)*100)], columns=['gain','base','lift'])
    df_gains.set_index('base', inplace=True)

    return df_gains, fig


def roc_curve(df, observed_column = 'observed', ranking_column = 'pred_ranking', title='ROC', xaxis_title='False positive rate',
                yaxis_title='True postitive rate'):

    fpr,tpr,_ = sklearn.metrics.roc_curve(df[observed_column], df[ranking_column])

    auc = sklearn.metrics.auc(fpr,tpr)

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=np.linspace(0,1,len(fpr)), y=np.linspace(0,1,len(fpr)), 
                            mode='lines', name='random'))
    fig.add_trace(go.Scatter(x=fpr, y=tpr,
                            mode='lines', name='ROC'))

    fig.update_layout(title=title, xaxis_title=xaxis_title,
                        yaxis_title=yaxis_title)

    return auc, fig
    

def rate_plot(df, thresholds = [0.5], observed_column = 'observed', pred_column = 'pred_observed', ranking_column = 'pred_ranking'):

    stats = []
    for t in thresholds:
        df[pred_column] = df[ranking_column].apply(lambda x: 1 if x > t else 0)

        accuracy = sklearn.metrics.accuracy_score(
            df[observed_column], df[pred_column])
        
        print("  threshold={}, accuracy={:.2f}".format(t, accuracy))
        
        tn, fp, fn, tp = sklearn.metrics.confusion_matrix(df[observed_column], 
                                                          df[pred_column]).ravel()
        
        total_positive = fn+tp
        total_negative = fp+tn
        total_pred_positive = df[pred_column].sum()
        response_rate = total_positive / len(df)
        pred_response_rate = total_pred_positive / len(df)
        precision_1 = tp/(fp+tp)
        recall_1 = tp/total_positive
        precision_0 = tn/(fn+tn)
        recall_0 = tn/total_negative

        stats.append((t, accuracy, response_rate, pred_response_rate, precision_1, recall_1, precision_0, recall_0))

    df_stats = pd.DataFrame(stats, columns=['threshold', 'accuracy', 'resp_rate', 'pred_resp_rate', 'precision_1', 'recall_1', 'precision_0', 'recall_0'])
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=thresholds, y=df_stats['accuracy'],
                            mode='lines', name='accuracy'))
    fig.add_trace(go.Scatter(x=thresholds, y=df_stats['resp_rate'],
                            mode='lines', name='resp rate'))
    fig.add_trace(go.Scatter(x=thresholds, y=df_stats['pred_resp_rate'],
                            mode='lines', name='pred resp rate'))
    fig.add_trace(go.Scatter(x=thresholds, y=df_stats['precision_1'],
                            mode='lines', name='precision_1'))
    fig.add_trace(go.Scatter(x=thresholds, y=df_stats['recall_1'],
                            mode='lines', name='recall_1'))      

    return df_stats, fig




import sys
import pandas as pd

import requests
import rauth
from rauth import OAuth2Service
import json

class MailchimpREST:
    
    def __init__(self, access_token = None, client_id = '725681533252', client_secret = '68e3936817108f1e9ce8fe7c4021a05a8d0cffddb5e3d82c0a'):
        
        self.access_token = access_token
        
        self.oauth_base_uri = 'https://login.mailchimp.com/oauth2/'
        self.authorize_uri = 'https://login.mailchimp.com/oauth2/authorize'
        self.access_token_uri = 'https://login.mailchimp.com/oauth2/token'
        self.base_uri = ''

        self.client_id=client_id
        self.client_secret=client_secret
        self.redirect_uri='http://127.0.0.1'
        
        
    def get_auth_url(self, client_id = None, client_secret = None):
        
        if client_id is not None:
            self.client_id = client_id
            
        if client_secret is not None:
            self.client_secret = client_secret
        
        oauth2 = OAuth2Service(
            client_id=self.client_id,
            client_secret=self.client_secret,
            name='Mailchimp',
            authorize_url=self.authorize_uri,
            access_token_url=self.access_token_uri,
            base_url=self.base_uri)

        params = {
            'response_type': 'code',
            'redirect_uri': self.redirect_uri
        }

        url = oauth2.get_authorize_url(**params)
        return url
    
    def get_access_token(self, code):
        
        data={
            'grant_type':'authorization_code',
            'client_id' : client_id,
            'client_secret' : client_secret,
            'redirect_uri' : redirect_uri,
            'code' : code
        }
        
        response = requests.post(access_token_uri, data=data)
        if response.status_code == 200:
            
            self.access_token = response.json()['access_token']
            print("[get_access_token] Success. Token = '{}'".format(self.access_token))
            
        else:
            print("[get_access_token] Failed.")
            print(response.json())
            
        return self.access_token
            
        #print(response.json())
        
    def _request(self, request_func, endpoint, params={}, data={}, json={}, verbose=False):
        
        if self.access_token == None or self.access_token == '':
            print("[_request] Access token '{}' is not valid.".format(self.access_token))
            return None
        
        if self.base_uri == '':
            response = requests.get(self.oauth_base_uri + 'metadata', headers={
                'Authorization':'OAuth {}'.format(self.access_token)
            })
            if response.status_code == 200:
                #print(response.json())
                self.base_uri = response.json()['api_endpoint'] + '/3.0/'                
            else:
                print("[_request] Could not retrieve metadata.")
                return None
        
        url = self.base_uri + endpoint
        response = request_func(url, params=params, data=data, json=json, headers={
            'Authorization':'OAuth {}'.format(self.access_token)
        })

        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {0}".format(response.status_code))
            print("Response: " + response.text)

        return response, response.json()
            
        
    def post(self, endpoint, params={}, data={}, json={}, verbose=False):

        return self._request(requests.post, endpoint, params=params, data=data, json=json, verbose=verbose)

    
    def delete(self, endpoint, verbose=False):
        
        return self._request(requests.delete, endpoint, verbose=verbose)


    def put(self, endpoint, params={}, data={}, json={}, verbose=False):
        
        return self._request(requests.put, endpoint, params=params, data=data, json=json, verbose=verbose)

    
    def patch(self, endpoint, params={}, data={}, json={}, verbose=False):
        
        return self._request(requests.patch, endpoint, params=params, data=data, json=json, verbose=verbose)

    
    def get(self, endpoint, params={}, verbose=False):
        
        return self._request(requests.get, endpoint, params=params, verbose=verbose)


class MailchimpCustomersExtract():
    
    """
    Wrapper class for extracting customers from a Mailchimp audience.
    """
    
    def __init__(self, access_token = '1976af7eb0470a7d844ec590697f37d5'):
        
        self.api = MailchimpREST(access_token)
        self.df_customers = None
        self.list_id = ''
        
    def _extract_field(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else ''
    
    def _extract_array(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else ['']
    
    def _extract_dict(self, customer, field_name):
        
        return customer[field_name] if field_name in customer else {}
    
    def _get_audience_list_id(self, audience='HUMAI'):
        
        response, data = self.api.get('lists', verbose=False)
        for l in data['lists']:
            if l['name'] == audience:
                self.list_id = l['id']
                return True
                
        return False
    
    def extract(self, audience='HUMAI', page_size = 1000, verbose=False):
        
        if not self._get_audience_list_id(audience=audience):
            print("[extract] Could not find list ''".format(audience))
            return None
    
        offset = 0
        total_members = 1
        page_size = 1000
        customers = []
        while offset < total_members:

            try:
                response, data = self.api.get('lists/{}/members'.format(self.list_id), params={
                    'count' : page_size,
                    'offset' : offset,

                    'fields' : 'merge_fields,total_items,members'
                })

                total_members = data['total_items']
                if offset % 5000 == 0:
                    print("{:,}/{:,} ({:.2f}%)".format(offset, total_members, 100.0*offset/total_members))

                if response.status_code == 200:

                    for member in data['members']:

                        fields = self._extract_dict(member, 'merge_fields')
                        address = self._extract_dict(fields, 'ADDRESS')
                        if isinstance(address, str):
                            address = {}
                            
                        customers.append( 
                            (
                                self._extract_field(member, 'id'),
                                (self._extract_field(fields,'FNAME') + ' ' + self._extract_field(fields,'LNAME')).strip(),
                                self._extract_field(member, 'email_address').strip(),
                                self._extract_field(fields, 'BIRTHDAY'),
                                self._extract_field(address, 'addr1').strip().upper(),
                                self._extract_field(address, 'city').strip().upper(),
                                self._extract_field(address, 'zip').strip().upper(),
                                self._extract_field(address, 'country').strip().upper()
                            )
                        )
                else:
                    print("[extract] Got response status {} ({}), skipping to next offset {:,}".format(response.status_code, response.text, offset+page_size))
            except:
                print("[extract] Unexpected error: " + sys.exc_info()[0])


            offset = page_size + offset

        print("{:,}/{:,} ({:.2f}%)".format(total_members, total_members, 100.0))
        
        self.df_customers = pd.DataFrame(customers, columns=[
            'Customer_ID', 'Name', 'Email', 'Birthday', 'Address', 'City', 'Zip', 'Country' 
        ])

        return self.df_customers


        
class MailchimpPredictionsLoader:
    
    """
    Helper class for pushing predictions to Mailchimp.
    
    TODO: Finish implementation.
    """
    
    def __init__(self, audience='HUMAI', access_token = '1976af7eb0470a7d844ec590697f37d5'):
        
        self.api = MailchimpREST(access_token)
        self.ready = False
        self.audience = audience
        self.list_id = ''
        self.field_to_tag = {}
        
    def create_or_init(self, audience='HUMAI'):
        
        self.audience = audience
        self.field_to_tag = {}
        self.list_id = ''
        self.ready = False
        self._create_fields(audience)
        
        
    def _get_audience_list_id(self, audience='HUMAI'):
        
        response, data = self.api.get('lists', verbose=False)
        for l in data['lists']:
            if l['name'] == audience:
                self.list_id = l['id']
                return True
                
        return False
    
        
    def _create_fields(self, audience):
        
        try:
        
            self.ready = False
            if not self._get_audience_list_id(audience):
                print("[_create_fields] Could not find audience ''".format(audience))
                return self.ready

            required_fields = [
                ('Date to email', 'date', True, ''),
                ('Time of day', 'text', True, ''),
                ('Propensity to Open', 'number', True, 0.0),
                ('Propensity to Buy', 'number', True, 0.0),
                ('Num Transactions', 'number', True, 0),
                ('Discount Sensitivity', 'number', True, 0),
                ('Life Cycle Stage', 'text', True, ''),
                ('Last Transaction Date', 'date', True, ''),
                ('Last Email Activity Date', 'date', True, '')
            ]

            response, data = self.api.get('lists/{}/merge-fields'.format(self.list_id))

            self.field_to_tag = {e['name']:e['tag'] for e in data['merge_fields'] }
            print(self.field_to_tag.keys())
                  
            for field in required_fields:
                if field[0] not in self.field_to_tag:
                    
                    print("Creating field '{}'".format(field[0]))
                    
                    response, data = self.api.post('lists/{}/merge-fields'.format(self.list_id), json={
                    #response, data = self.api.patch('lists/{}/merge-fields'.format(self.list_id), json={
                        'name' : field[0],
                        'type' : field[1],
                        'public' : field[2]
                    }, verbose=False)
                    
                    #data = response
                    #print(data)
                    
                    self.field_to_tag[field[0]] = data['tag']
        
            self.ready = True
            
            #print(self.field_to_tag)
            
        except:
            
            print("[extract] Unexpected error: " + sys.exc_info()[0])
            
        return self.ready
        
    
    def push(
        self, 
        df,
        mappings,
        verbose=False):

        if not self.ready:
            print("[push] Not ready, call create_or_init()")
            return False
        
        nn = len(df)
        ii = 0

        dfg = df.groupby(by='Customer_ID')
        for cid, df_predictions in dfg:
            
            if ii % 500 == 0:
                print("{:,}/{:,} ({:.2f} %)".format(ii, nn, 100.0*ii/nn))
            
            if cid == '':
                print("  Skipping invalid Customer ID: {} with predictions {}".format(copernica_id, df_predictions))
                ii = len(df_predictions) + ii
            else:
                
                for index, row in df_predictions.iterrows():
                    
                    #print(row)
                    
                    merge_fields = {
                        self.field_to_tag[field] : row[mappings[field]] for field in mappings.keys()
                    }
                    
                    response = self.api.patch('lists/{}/members/{}'.format(self.list_id, cid), json={
                        'merge_fields' : merge_fields
                    }, verbose=False)


            ii = 1 + ii
            
        print("{:,}/{:,} ({:.2f} %)".format(nn, nn, 100.0))
        
        return True
        

    def remove(self, df, pred_type = 'PersonalizedMailFlow', verbose=False):

        if not self.ready:
            print("[push_predictions_to_copernica] Not ready, call create_or_init()")
            return False
        
        assert False, "Implement."
        
        return False 